import einops
import numpy as np
import torch
from eindex import eindex
from jaxtyping import Float, Int
from sae_lens import SAE
from torch import Tensor
from transformer_lens import HookedTransformer, utils

from sae_dashboard.components import LogitsTableData, SequenceData
from sae_dashboard.sae_vis_data import SaeVisData
from sae_dashboard.transformer_lens_wrapper import (
    ActivationConfig,
    TransformerLensWrapper,
    to_resid_direction,
)
from sae_dashboard.utils_fns import RollingCorrCoef, TopK

Arr = np.ndarray


def get_features_table_data(
    feature_out_dir: Float[Tensor, "feats d_out"],
    n_rows: int,
    corrcoef_neurons: RollingCorrCoef | None = None,
    corrcoef_encoder: RollingCorrCoef | None = None,
) -> dict[str, list[list[int]] | list[list[float]]]:
    # ! Calculate all data for the left-hand column visualisations, i.e. the 3 tables
    # Store kwargs (makes it easier to turn the tables on and off individually)
    feature_tables_data: dict[str, list[list[int]] | list[list[float]]] = {}

    # Table 1: neuron alignment, based on decoder weights
    # if layout.feature_tables_cfg.neuron_alignment_table:
    # Let's just always do this.
    add_neuron_alignment_data(
        feature_out_dir=feature_out_dir,
        feature_tables_data=feature_tables_data,
        n_rows=n_rows,
    )

    # Table 2: neurons correlated with this feature, based on their activations
    if corrcoef_neurons is not None:
        add_feature_neuron_correlations(
            corrcoef_neurons=corrcoef_neurons,
            feature_tables_data=feature_tables_data,
            n_rows=n_rows,
        )

    # Table 3: primary encoder features correlated with this feature, based on their activations
    if corrcoef_encoder is not None:
        add_intra_encoder_correlations(
            corrcoef_encoder=corrcoef_encoder,
            feature_tables_data=feature_tables_data,
            n_rows=n_rows,
        )

    return feature_tables_data


def add_intra_encoder_correlations(
    corrcoef_encoder: RollingCorrCoef,
    feature_tables_data: dict[str, list[list[int]] | list[list[float]]],
    n_rows: int,
):
    enc_indices, enc_pearson, enc_cossim = corrcoef_encoder.topk_pearson(
        k=n_rows,
    )
    feature_tables_data["correlated_features_indices"] = enc_indices
    feature_tables_data["correlated_features_pearson"] = enc_pearson
    feature_tables_data["correlated_features_cossim"] = enc_cossim


def add_neuron_alignment_data(
    feature_out_dir: Float[Tensor, "feats d_out"],
    feature_tables_data: dict[str, list[list[int]] | list[list[float]]],
    n_rows: int,
):
    top3_neurons_aligned = TopK(tensor=feature_out_dir.float(), k=n_rows, largest=True)
    feature_out_l1_norm = feature_out_dir.abs().sum(dim=-1, keepdim=True)
    pct_of_l1: Arr = np.absolute(top3_neurons_aligned.values) / utils.to_numpy(
        feature_out_l1_norm.float()
    )
    feature_tables_data["neuron_alignment_indices"] = (
        top3_neurons_aligned.indices.tolist()
    )
    feature_tables_data["neuron_alignment_values"] = (
        top3_neurons_aligned.values.tolist()
    )
    feature_tables_data["neuron_alignment_l1"] = pct_of_l1.tolist()


def add_feature_neuron_correlations(
    corrcoef_neurons: RollingCorrCoef,
    feature_tables_data: dict[str, list[list[int]] | list[list[float]]],
    n_rows: int,
):
    neuron_indices, neuron_pearson, neuron_cossim = corrcoef_neurons.topk_pearson(
        k=n_rows,
    )

    feature_tables_data["correlated_neurons_indices"] = neuron_indices
    feature_tables_data["correlated_neurons_pearson"] = neuron_pearson
    feature_tables_data["correlated_neurons_cossim"] = neuron_cossim


def get_logits_table_data(logit_vector: Float[Tensor, "d_vocab"], n_rows: int):
    # Get logits table data
    top_logits = TopK(logit_vector.float(), k=n_rows, largest=True)
    bottom_logits = TopK(logit_vector.float(), k=n_rows, largest=False)

    top_logit_values = top_logits.values.tolist()
    top_token_ids = top_logits.indices.tolist()

    bottom_logit_values = bottom_logits.values.tolist()
    bottom_token_ids = bottom_logits.indices.tolist()

    logits_table_data = LogitsTableData(
        bottom_logits=bottom_logit_values,
        bottom_token_ids=bottom_token_ids,
        top_logits=top_logit_values,
        top_token_ids=top_token_ids,
    )

    return logits_table_data


# @torch.inference_mode()
# def get_feature_data(
#     encoder: AutoEncoder,
#     model: HookedTransformer,
#     tokens: Int[Tensor, "batch seq"],
#     cfg: SaeVisConfig,
# ) -> SaeVisData:
#     """
#     This is the main function which users will run to generate the feature visualization data. It batches this
#     computation over features, in accordance with the arguments in the SaeVisConfig object (we don't want to compute all
#     the features at once, since might give OOMs).

#     See the `_get_feature_data` function for an explanation of the arguments, as well as a more detailed explanation
#     of what this function is doing.

#     The return object is the merged SaeVisData objects returned by the `_get_feature_data` function.
#     """
#     pass

#     # return sae_vis_data


@torch.inference_mode()
def parse_prompt_data(
    tokens: Int[Tensor, "batch seq"],
    str_toks: list[str],
    sae_vis_data: SaeVisData,
    feat_acts: Float[Tensor, "seq feats"],
    feature_resid_dir: Float[Tensor, "feats d_model"],
    resid_post: Float[Tensor, "seq d_model"],
    W_U: Float[Tensor, "d_model d_vocab"],
    feature_idx: list[int] | None = None,
    num_top_features: int = 10,
) -> dict[str, tuple[list[int], list[str]]]:
    """
    Gets data needed to create the sequences in the prompt-centric vis (displaying dashboards for the most relevant
    features on a prompt).

    This function exists so that prompt dashboards can be generated without using our AutoEncoder or
    TransformerLens(Wrapper) classes.

    Args:
        tokens: Int[Tensor, "batch seq"]
            The tokens we'll be using to get the feature activations. Note that we might not be using all of them; the
            number used is determined by `fvp.total_batch_size`.

        str_toks:  list[str]
            The tokens as a list of strings, so that they can be visualized in HTML.

        sae_vis_data: SaeVisData
             The object storing all data for each feature. We'll set each `feature_data.prompt_data` to the
             data we get from `prompt`.

        feat_acts: Float[Tensor, "seq feats"]
            The activations values of the features across the sequence.

        feature_resid_dir: Float[Tensor, "feats d_model"]
            The directions that each feature writes to the residual stream.

        resid_post: Float[Tensor, "seq d_model"]
            The activations of the final layer of the model before the unembed.

        W_U: Float[Tensor, "d_model d_vocab"]
            The model's unembed weights for the logit lens.

        feature_idx: list[int] or None
            The features we're actually computing. These might just be a subset of the model's full features.

        num_top_features: int
            The number of top features to display in this view, for any given metric.

    Returns:
        scores_dict: dict[str, tuple[list[int], list[str]]]
            A dictionary mapping keys like "act_quantile|'django' (0)" to a tuple of lists, where the first list is the
            feature indices, and the second list is the string-formatted values of the scores.

    As well as returning this dictionary, this function will also set `FeatureData.prompt_data` for each feature in
    `sae_vis_data` (this is necessary for getting the prompts in the prompt-centric vis). Note this design choice could
    have been done differently (i.e. have this function return a list of the prompt data for each feature). I chose this
    way because it means the FeatureData._get_html_data_prompt_centric can work fundamentally the same way as
    FeatureData._get_html_data_feature_centric, rather than treating the prompt data object as a different kind of
    component in the vis.
    """

    device = sae_vis_data.cfg.device

    if feature_idx is None:
        feature_idx = list(sae_vis_data.feature_data_dict.keys())
    n_feats = len(feature_idx)
    assert (
        feature_resid_dir.shape[0] == n_feats
    ), f"The number of features in feature_resid_dir ({feature_resid_dir.shape[0]}) does not match the number of feature indices ({n_feats})"

    assert (
        feat_acts.shape[1] == n_feats
    ), f"The number of features in feat_acts ({feat_acts.shape[1]}) does not match the number of feature indices ({n_feats})"

    feats_loss_contribution = torch.empty(
        size=(n_feats, tokens.shape[1] - 1), device=device
    )
    # Some logit computations which we only need to do once
    # correct_token_unembeddings = model_wrapped.W_U[:, tokens[0, 1:]] # [d_model seq]
    orig_logits = (
        resid_post / resid_post.std(dim=-1, keepdim=True)
    ) @ W_U  # [seq d_vocab]
    raw_logits = feature_resid_dir @ W_U  # [feats d_vocab]

    for i, feat in enumerate(feature_idx):
        # ! Calculate the sequence data for each feature, and store it as FeatureData.prompt_data

        # Get this feature's output vector, using an outer product over the feature activations for all tokens
        resid_post_feature_effect = einops.einsum(
            feat_acts[:, i], feature_resid_dir[i], "seq, d_model -> seq d_model"
        )

        # Ablate the output vector from the residual stream, and get logits post-ablation
        new_resid_post = resid_post - resid_post_feature_effect
        new_logits = (new_resid_post / new_resid_post.std(dim=-1, keepdim=True)) @ W_U

        # Get the top5 & bottom5 changes in logits (don't bother with `efficient_topk` cause it's small)
        contribution_to_logprobs = orig_logits.log_softmax(
            dim=-1
        ) - new_logits.log_softmax(dim=-1)
        top_contribution_to_logits = TopK(contribution_to_logprobs[:-1], k=5)
        bottom_contribution_to_logits = TopK(
            contribution_to_logprobs[:-1], k=5, largest=False
        )

        # Get the change in loss (which is negative of change of logprobs for correct token)
        loss_contribution = eindex(
            -contribution_to_logprobs[:-1], tokens[0, 1:], "seq [seq]"
        )
        feats_loss_contribution[i, :] = loss_contribution

        # Store the sequence data
        sae_vis_data.feature_data_dict[feat].prompt_data = SequenceData(
            token_ids=tokens.squeeze(0).tolist(),
            feat_acts=[round(f, 4) for f in feat_acts[:, i].tolist()],
            loss_contribution=[0.0] + loss_contribution.tolist(),
            token_logits=raw_logits[i, tokens.squeeze(0)].tolist(),
            top_token_ids=top_contribution_to_logits.indices.tolist(),
            top_logits=top_contribution_to_logits.values.tolist(),
            bottom_token_ids=bottom_contribution_to_logits.indices.tolist(),
            bottom_logits=bottom_contribution_to_logits.values.tolist(),
        )

    # ! Lastly, return a dictionary mapping each key like 'act_quantile|"django" (0)' to a list of feature indices & scores

    # Get a dict with keys like f"act_quantile|'My' (1)" and values (feature indices list, feature score values list)
    scores_dict: dict[str, tuple[list[int], list[str]]] = {}

    for seq_pos, seq_key in enumerate([f"{t!r} ({i})" for i, t in enumerate(str_toks)]):
        # Filter the feature activations, since we only need the ones that are non-zero
        feat_acts_nonzero_filter = utils.to_numpy(feat_acts[seq_pos] > 0)
        feat_acts_nonzero_locations = np.nonzero(feat_acts_nonzero_filter)[0].tolist()
        _feat_acts = feat_acts[seq_pos, feat_acts_nonzero_filter]  # [feats_filtered,]
        _feature_idx = np.array(feature_idx)[feat_acts_nonzero_filter]

        if feat_acts_nonzero_filter.sum() > 0:
            k = min(num_top_features, _feat_acts.numel())

            # Get the top features by activation size. This is just applying a TopK function to the feat acts (which
            # were stored by the code before this). The feat acts are formatted to 3dp.
            act_size_topk = TopK(_feat_acts, k=k, largest=True)
            top_features = _feature_idx[act_size_topk.indices].tolist()
            formatted_scores = [f"{v:.3f}" for v in act_size_topk.values]
            scores_dict[f"act_size|{seq_key}"] = (top_features, formatted_scores)

            # Get the top features by activation quantile. We do this using the `feature_act_quantiles` object, which
            # was stored `sae_vis_data`. This quantiles object has a method to return quantiles for a given set of
            # data, as well as the precision (we make the precision higher for quantiles closer to 100%, because these
            # are usually the quantiles we're interested in, and it lets us to save space in `feature_act_quantiles`).
            act_quantile, act_precision = sae_vis_data.feature_stats.get_quantile(
                _feat_acts, feat_acts_nonzero_locations
            )
            act_quantile_topk = TopK(act_quantile, k=k, largest=True)
            act_formatting = [
                f".{act_precision[i]-2}%" for i in act_quantile_topk.indices
            ]
            top_features = _feature_idx[act_quantile_topk.indices].tolist()
            formatted_scores = [
                f"{v:{f}}" for v, f in zip(act_quantile_topk.values, act_formatting)
            ]
            scores_dict[f"act_quantile|{seq_key}"] = (top_features, formatted_scores)

        # We don't measure loss effect on the first token
        if seq_pos == 0:
            continue

        # Filter the loss effects, since we only need the ones which have non-zero feature acts on the tokens before them
        prev_feat_acts_nonzero_filter = utils.to_numpy(feat_acts[seq_pos - 1] > 0)
        _loss_contribution = feats_loss_contribution[
            prev_feat_acts_nonzero_filter, seq_pos - 1
        ]  # [feats_filtered,]
        _feature_idx_prev = np.array(feature_idx)[prev_feat_acts_nonzero_filter]

        if prev_feat_acts_nonzero_filter.sum() > 0:
            k = min(num_top_features, _loss_contribution.numel())

            # Get the top features by loss effect. This is just applying a TopK function to the loss effects (which were
            # stored by the code before this). The loss effects are formatted to 3dp. We look for the most negative
            # values, i.e. the most loss-reducing features.
            loss_contribution_topk = TopK(_loss_contribution, k=k, largest=False)
            top_features = _feature_idx_prev[loss_contribution_topk.indices].tolist()
            formatted_scores = [f"{v:+.3f}" for v in loss_contribution_topk.values]
            scores_dict[f"loss_effect|{seq_key}"] = (top_features, formatted_scores)
    return scores_dict


@torch.inference_mode()
def get_prompt_data(
    sae_vis_data: SaeVisData,
    prompt: str,
    num_top_features: int,
) -> dict[str, tuple[list[int], list[str]]]:
    """
    Gets data that will be used to create the sequences in the prompt-centric HTML visualisation, i.e. an object of
    type SequenceData for each of our features.

    Args:
        sae_vis_data:     The object storing all data for each feature. We'll set each `feature_data.prompt_data` to the
                          data we get from `prompt`.
        prompt:           The prompt we'll be using to get the feature activations.#
        num_top_features: The number of top features we'll be getting data for.

    Returns:
        scores_dict:      A dictionary mapping keys like "act_quantile|0" to a tuple of lists, where the first list is
                          the feature indices, and the second list is the string-formatted values of the scores.

    As well as returning this dictionary, this function will also set `FeatureData.prompt_data` for each feature in
    `sae_vis_data`. This is because the prompt-centric vis will call `FeatureData._get_html_data_prompt_centric` on each
    feature data object, so it's useful to have all the data in once place! Even if this will get overwritten next
    time we call `get_prompt_data` for this same `sae_vis_data` object.
    """

    # ! Boring setup code
    feature_idx = list(sae_vis_data.feature_data_dict.keys())
    encoder = sae_vis_data.encoder
    assert isinstance(encoder, SAE)
    model = sae_vis_data.model
    assert isinstance(model, HookedTransformer)
    cfg = sae_vis_data.cfg
    assert isinstance(cfg.hook_point, str), f"{cfg.hook_point=}, expected a string"

    str_toks: list[str] = model.tokenizer.tokenize(prompt)  # type: ignore
    tokens = model.tokenizer.encode(prompt, return_tensors="pt").to(  # type: ignore
        sae_vis_data.cfg.device
    )
    assert isinstance(tokens, torch.Tensor)

    model_wrapped = TransformerLensWrapper(model, ActivationConfig(cfg.hook_point, []))

    feature_act_dir = encoder.W_enc[:, feature_idx]  # [d_in feats]
    feature_out_dir = encoder.W_dec[feature_idx]  # [feats d_in]
    feature_resid_dir = to_resid_direction(
        feature_out_dir, model_wrapped
    )  # [feats d_model]
    assert (
        feature_act_dir.T.shape
        == feature_out_dir.shape
        == (len(feature_idx), encoder.cfg.d_in)
    )

    # ! Define hook functions to cache all the info required for feature ablation, then run those hook fns
    resid_post, act_post = model_wrapped(tokens, return_logits=False)
    resid_post: Tensor = resid_post.squeeze(0)
    feat_acts = encoder.get_feature_acts_subset(act_post, feature_idx).squeeze(
        0
    )  # [seq feats]

    # ! Use the data we've collected to make the scores_dict and update the sae_vis_data
    scores_dict = parse_prompt_data(
        tokens=tokens,
        str_toks=str_toks,
        sae_vis_data=sae_vis_data,
        feat_acts=feat_acts,
        feature_resid_dir=feature_resid_dir,
        resid_post=resid_post,
        W_U=model.W_U,
        feature_idx=feature_idx,
        num_top_features=num_top_features,
    )

    return scores_dict



from typing import Any, Dict, List, Union

import einops
import numpy as np
import torch
from sae_lens import SAE
from transformer_lens import ActivationCache, HookedTransformer


class DFACalculator:
    """Calculate DFA values for a given layer and set of feature indices."""

    def __init__(self, model: HookedTransformer, sae: SAE):
        self.model = model
        self.sae = sae
        if (
            hasattr(model.cfg, "n_key_value_heads")
            and model.cfg.n_key_value_heads is not None
            and model.cfg.n_key_value_heads < model.cfg.n_heads
        ):
            print("Using GQA")
            self.use_gqa = True
        else:
            self.use_gqa = False

    def calculate(
        self,
        activations: Union[Dict[str, torch.Tensor], ActivationCache],
        layer_num: int,
        feature_indices: List[int],
        max_value_indices: torch.Tensor,
    ) -> Dict[int, Any]:  # type: ignore
        """Calculate DFA values for a given layer and set of feature indices."""
        if not feature_indices:
            return {}

        v = activations[f"blocks.{layer_num}.attn.hook_v"]
        attn_weights = activations[f"blocks.{layer_num}.attn.hook_pattern"]

        if self.use_gqa:
            per_src_pos_dfa = self.calculate_gqa_intermediate_tensor(
                attn_weights, v, feature_indices
            )
        else:
            per_src_pos_dfa = self.calculate_standard_intermediate_tensor(
                attn_weights, v, feature_indices
            )

        n_prompts, seq_len, _, n_features = per_src_pos_dfa.shape

        # Use advanced indexing to get per_src_dfa
        prompt_indices = torch.arange(n_prompts)[:, None, None]
        src_pos_indices = torch.arange(seq_len)[None, :, None]
        feature_indices_tensor = torch.arange(n_features)[None, None, :]
        max_value_indices_expanded = max_value_indices[:, None, :]

        per_src_dfa = per_src_pos_dfa[
            prompt_indices,
            max_value_indices_expanded,
            src_pos_indices,
            feature_indices_tensor,
        ]

        max_values, _ = per_src_dfa.max(dim=1)

        # Create a structured numpy array to hold all the data
        dtype = np.dtype(
            [
                ("dfa_values", np.float32, (seq_len,)),
                ("dfa_target_index", np.int32),
                ("dfa_max_value", np.float32),
            ]
        )
        results = np.zeros((len(feature_indices), n_prompts), dtype=dtype)

        # Fill the numpy array with data
        results["dfa_values"] = per_src_dfa.detach().cpu().numpy().transpose(2, 0, 1)
        results["dfa_target_index"] = max_value_indices.detach().cpu().numpy().T
        results["dfa_max_value"] = max_values.detach().cpu().numpy().T

        # Create a dictionary mapping feature indices to their respective data
        final_results = {
            feat_idx: results[i] for i, feat_idx in enumerate(feature_indices)
        }

        return final_results

    def calculate_standard_intermediate_tensor(
        self,
        attn_weights: torch.Tensor,
        v: torch.Tensor,
        feature_indices: List[int],
    ) -> torch.Tensor:
        v_cat = einops.rearrange(
            v, "batch src_pos n_heads d_head -> batch src_pos (n_heads d_head)"
        )

        attn_weights_bcast = einops.repeat(
            attn_weights,
            "batch n_heads dest_pos src_pos -> batch dest_pos src_pos (n_heads d_head)",
            d_head=self.model.cfg.d_head,
        )

        decomposed_z_cat = attn_weights_bcast * v_cat.unsqueeze(1)

        W_enc_selected = self.sae.W_enc[:, feature_indices]  # [d_model, num_indices]

        per_src_pos_dfa = einops.einsum(
            decomposed_z_cat,
            W_enc_selected,
            "batch dest_pos src_pos d_model, d_model num_features -> batch dest_pos src_pos num_features",
        )

        return per_src_pos_dfa

    def calculate_gqa_intermediate_tensor(
        self, attn_weights: torch.Tensor, v: torch.Tensor, feature_indices: List[int]
    ) -> torch.Tensor:

        n_query_heads = attn_weights.shape[1]
        n_kv_heads = v.shape[2]
        expansion_factor = n_query_heads // n_kv_heads
        v = v.repeat_interleave(expansion_factor, dim=2)

        v_cat = einops.rearrange(
            v, "batch src_pos n_heads d_head -> batch src_pos (n_heads d_head)"
        )

        W_enc_selected = self.sae.W_enc[:, feature_indices]  # [d_model, num_indices]

        # Initialize the result tensor
        n_prompts, seq_len, _ = v_cat.shape
        n_features = len(feature_indices)
        per_src_pos_dfa = torch.zeros(
            (n_prompts, seq_len, seq_len, n_features), device=v_cat.device
        )

        # Process in chunks
        chunk_size = 16  # Adjust this based on your memory constraints
        for i in range(0, seq_len, chunk_size):
            chunk_end = min(i + chunk_size, seq_len)

            # Process a chunk of destination positions
            attn_weights_chunk = attn_weights[:, :, i:chunk_end, :]
            attn_weights_bcast_chunk = einops.repeat(
                attn_weights_chunk,
                "batch n_heads dest_pos src_pos -> batch dest_pos src_pos (n_heads d_head)",
                d_head=self.model.cfg.d_head,
            )
            decomposed_z_cat_chunk = attn_weights_bcast_chunk * v_cat.unsqueeze(1)

            per_src_pos_dfa_chunk = einops.einsum(
                decomposed_z_cat_chunk,
                W_enc_selected,
                "batch dest_pos src_pos d_model, d_model num_features -> batch dest_pos src_pos num_features",
            )

            per_src_pos_dfa[:, i:chunk_end, :, :] = per_src_pos_dfa_chunk

        return per_src_pos_dfa



from dataclasses import dataclass
from typing import List, Optional

DEFAULT_SPARSITY_THRESHOLD = -6


@dataclass
class NeuronpediaRunnerConfig:

    sae_set: str
    sae_path: str
    outputs_dir: str
    np_sae_id_suffix: Optional[str] = (
        None  # this is the __[np_sae_id_suffix] after the SAE Set
    )
    np_set_name: Optional[str] = None
    from_local_sae: bool = False
    sparsity_threshold: int = DEFAULT_SPARSITY_THRESHOLD
    huggingface_dataset_path: str = ""

    # ACTIVATION STORE PARAMETERS
    # token pars
    n_prompts_total: int = 24576
    n_tokens_in_prompt: int = 128
    n_prompts_in_forward_pass: int = 32

    # batching
    n_features_at_a_time: int = 128
    quantile_feature_batch_size: int = 64
    start_batch: int = 0
    end_batch: Optional[int] = None

    # quantiles
    n_quantiles: int = 5
    top_acts_group_size: int = 30
    quantile_group_size: int = 5

    # additional calculations
    use_dfa: bool = False

    model_dtype: str = ""
    sae_dtype: str = ""
    model_id: Optional[str] = None
    layer: Optional[int] = None

    sae_device: str | None = None
    activation_store_device: str | None = None
    model_device: str | None = None
    model_n_devices: int | None = None
    use_wandb: bool = False

    shuffle_tokens: bool = True
    prefix_tokens: Optional[List[int]] = None
    suffix_tokens: Optional[List[int]] = None
    ignore_positions: Optional[List[int]] = None



from dataclasses import asdict, dataclass, field

from dataclasses_json import dataclass_json
from rich import print as rprint
from rich.tree import Tree

from sae_dashboard.components_config import (
    ActsHistogramConfig,
    BaseComponentConfig,
    Column,
    FeatureTablesConfig,
    LogitsHistogramConfig,
    LogitsTableConfig,
    PromptConfig,
    SequencesConfig,
)

KEY_LAYOUT_VIS = """Key:
  the tree shows which components will be displayed in each column (from left to right)
  arguments are [b #00aa00]green[/]
  arguments changed from their default are [b dark_orange]orange[/], with default in brackets
  argument descriptions are in [i]italics[/i]
"""


@dataclass_json
@dataclass
class SaeVisLayoutConfig:
    """
    This object allows you to set all the ways the feature vis will be laid out.

    Args (specified by the user):
        columns:
            A list of `Column` objects, where each `Column` contains a list of component configs.
        height:
            The height of the vis (in pixels).

    Args (defined during __init__):
        seq_cfg:
            The `SequencesConfig` object, which contains all the parameters for the top activating sequences (and the
            quantile groups).
        act_hist_cfg:
            The `ActsHistogramConfig` object, which contains all the parameters for the activations histogram.
        logits_hist_cfg:
            The `LogitsHistogramConfig` object, which contains all the parameters for the logits histogram.
        logits_table_cfg:
            The `LogitsTableConfig` object, which contains all the parameters for the logits table.
        feature_tables_cfg:
            The `FeatureTablesConfig` object, which contains all the parameters for the feature tables.
        prompt_cfg:
            The `PromptConfig` object, which contains all the parameters for the prompt-centric vis.
    """

    columns: dict[int | tuple[int, int], Column] = field(default_factory=dict)
    height: int = 750

    seq_cfg: SequencesConfig | None = None
    act_hist_cfg: ActsHistogramConfig | None = None
    logits_hist_cfg: LogitsHistogramConfig | None = None
    logits_table_cfg: LogitsTableConfig | None = None
    feature_tables_cfg: FeatureTablesConfig | None = None
    prompt_cfg: PromptConfig | None = None

    def __init__(self, columns: list[Column], height: int = 750):
        """
        The __init__ method will allow you to extract things like `self.seq_cfg` from the object (even though they're
        initially stored in the `columns` attribute). It also verifies that there are no duplicate components (which is
        redundant, and could mess up the HTML).
        """
        # Define the columns (as dict) and the height
        self.columns = {idx: col for idx, col in enumerate(columns)}
        self.height = height

        # Get a list of all our components, and verify there's no duplicates
        all_components = [
            component for column in self.columns.values() for component in column
        ]
        all_component_names = [
            comp.__class__.__name__.rstrip("Config") for comp in all_components
        ]
        assert len(all_component_names) == len(
            set(all_component_names)
        ), "Duplicate components in layout config"
        self.components: dict[str, BaseComponentConfig] = {
            name: comp for name, comp in zip(all_component_names, all_components)
        }

        # Once we've verified this, store each config component as an attribute
        for comp, comp_name in zip(all_components, all_component_names):
            match comp_name:
                case "Prompt":
                    self.prompt_cfg = comp
                case "Sequences":
                    self.seq_cfg = comp
                case "ActsHistogram":
                    self.act_hist_cfg = comp
                case "LogitsHistogram":
                    self.logits_hist_cfg = comp
                case "LogitsTable":
                    self.logits_table_cfg = comp
                case "FeatureTables":
                    self.feature_tables_cfg = comp
                case _:
                    raise ValueError(f"Unknown component name {comp_name}")

    def data_is_contained_in(self, other: "SaeVisLayoutConfig") -> bool:
        """
        Returns True if `self` uses only data that would already exist in `other`. This is useful because our prompt-
        centric vis needs to only use data that was already computed as part of our initial data gathering. For example,
        if our SaeVisData object only contains the first 10 rows of the logits table, then we can't show the top 15 rows
        in the prompt centric view!
        """
        for comp_name, comp in self.components.items():
            # If the component in `self` is not present in `other`, return False
            if comp_name not in other.components:
                return False
            # If the component in `self` is present in `other`, but the `self` component is larger, then return False
            comp_other = other.components[comp_name]
            if not comp.data_is_contained_in(comp_other):
                return False

        return True

    def help(
        self,
        title: str = "SaeVisLayoutConfig",
        key: bool = True,
    ) -> Tree | None:
        """
        This prints out a tree showing the layout of the vis, by column (as well as the values of the arguments for each
        config object, plus their default values if they changed, and the descriptions of each arg).
        """

        # Create tree (with title and optionally the key explaining arguments)
        if key:
            title += "\n\n" + KEY_LAYOUT_VIS
        tree = Tree(title)

        n_columns = len(self.columns)

        # For each column, add a tree node
        for column_idx, vis_components in self.columns.items():
            n_components = len(vis_components)
            tree_column = tree.add(f"Column {column_idx}")

            # For each component in that column, add a tree node
            for component_idx, vis_component in enumerate(vis_components):
                n_params = len(asdict(vis_component))
                tree_component = tree_column.add(
                    f"{vis_component.__class__.__name__}".rstrip("Config")
                )

                # For each config parameter of that component
                for param_idx, (param, value) in enumerate(
                    asdict(vis_component).items()
                ):
                    # Get line break if we're at the final parameter of this component (unless it's the final component
                    # in the final column)
                    suffix = "\n" if (param_idx == n_params - 1) else ""
                    if (component_idx == n_components - 1) and (
                        column_idx == n_columns - 1
                    ):
                        suffix = ""

                    # Get argument description, and its default value
                    desc = vis_component.help_dict.get(param, "")
                    value_default = getattr(
                        vis_component.__class__, param, "no default"
                    )

                    # Add tree node (appearance is different if value is changed from default)
                    if value != value_default:
                        info = f"[b dark_orange]{param}: {value!r}[/] ({value_default!r}) \n[i]{desc}[/i]{suffix}"
                    else:
                        info = (
                            f"[b #00aa00]{param}: {value!r}[/] \n[i]{desc}[/i]{suffix}"
                        )
                    tree_component.add(info)

        rprint(tree)

    @classmethod
    def default_feature_centric_layout(cls) -> "SaeVisLayoutConfig":
        return cls(
            columns=[
                Column(FeatureTablesConfig()),
                Column(
                    ActsHistogramConfig(), LogitsTableConfig(), LogitsHistogramConfig()
                ),
                Column(SequencesConfig(stack_mode="stack-none")),
            ],
            height=750,
        )

    @classmethod
    def default_prompt_centric_layout(cls) -> "SaeVisLayoutConfig":
        return cls(
            columns=[
                Column(
                    PromptConfig(),
                    ActsHistogramConfig(),
                    LogitsTableConfig(n_rows=5),
                    SequencesConfig(top_acts_group_size=10, n_quantiles=0),
                    width=450,
                ),
            ],
            height=1000,
        )



__version__ = "0.6.2"

# from .data_fetching_fns import *
# from .data_storing_fns import *
# from .html_fns import *
# from .transformer_lens_wrapper import *
# from .utils_fns import *


# from autoencoder import AutoEncoder, AutoEncoderConfig



from pathlib import Path
from typing import Dict, List

import einops
import numpy as np
import torch
from jaxtyping import Float, Int
from sae_lens import SAE
from sae_lens.config import DTYPE_MAP as DTYPES
from torch import Tensor, nn
from tqdm.auto import tqdm

from sae_dashboard.dfa_calculator import DFACalculator
from sae_dashboard.sae_vis_data import SaeVisConfig
from sae_dashboard.transformer_lens_wrapper import (
    TransformerLensWrapper,
    to_resid_direction,
)
from sae_dashboard.utils_fns import RollingCorrCoef

Arr = np.ndarray


class FeatureDataGenerator:
    def __init__(
        self,
        cfg: SaeVisConfig,
        tokens: Int[Tensor, "batch seq"],
        model: TransformerLensWrapper,
        encoder: SAE,
    ):
        self.cfg = cfg
        self.model = model
        self.encoder = encoder
        self.token_minibatches = self.batch_tokens(tokens)
        self.dfa_calculator = (
            DFACalculator(model.model, encoder) if cfg.use_dfa else None
        )

        if cfg.use_dfa:
            assert (
                "hook_z" in encoder.cfg.hook_name
            ), f"DFAs are only supported for hook_z, but got {encoder.cfg.hook_name}"

    @torch.inference_mode()
    def batch_tokens(
        self, tokens: Int[Tensor, "batch seq"]
    ) -> list[Int[Tensor, "batch seq"]]:
        # Get tokens into minibatches, for the fwd pass
        token_minibatches = (
            (tokens,)
            if self.cfg.minibatch_size_tokens is None
            else tokens.split(self.cfg.minibatch_size_tokens)
        )
        token_minibatches = [tok.to(self.cfg.device) for tok in token_minibatches]

        return token_minibatches

    @torch.inference_mode()
    def get_feature_data(  # type: ignore
        self,
        feature_indices: list[int],
        progress: list[tqdm] | None = None,  # type: ignore
    ):  # type: ignore
        # Create lists to store the feature activations & final values of the residual stream
        all_feat_acts = []
        all_dfa_results = {feature_idx: {} for feature_idx in feature_indices}
        total_prompts = 0

        # Create objects to store the data for computing rolling stats
        corrcoef_neurons = RollingCorrCoef()
        corrcoef_encoder = RollingCorrCoef(indices=feature_indices, with_self=True)

        # Get encoder & decoder directions
        feature_out_dir = self.encoder.W_dec[feature_indices]  # [feats d_autoencoder]
        feature_resid_dir = to_resid_direction(
            feature_out_dir, self.model
        )  # [feats d_model]

        # ! Compute & concatenate together all feature activations & post-activation function values
        for i, minibatch in enumerate(self.token_minibatches):
            minibatch.to(self.cfg.device)
            model_activation_dict = self.get_model_acts(i, minibatch)
            primary_acts = model_activation_dict[
                self.model.activation_config.primary_hook_point
            ].to(
                self.encoder.device
            )  # make sure acts are on the correct device

            # Compute feature activations from this
            with FeatureMaskingContext(self.encoder, feature_indices):
                feature_acts = self.encoder.encode(primary_acts).to(
                    DTYPES[self.cfg.dtype]
                )

            self.update_rolling_coefficients(
                model_acts=primary_acts,
                feature_acts=feature_acts,
                corrcoef_neurons=corrcoef_neurons,
                corrcoef_encoder=corrcoef_encoder,
            )

            # Add these to the lists (we'll eventually concat)
            all_feat_acts.append(feature_acts)

            # Calculate DFA
            if self.cfg.use_dfa and self.dfa_calculator:
                max_value_indices = torch.argmax(feature_acts, dim=1)
                batch_dfa_results = self.dfa_calculator.calculate(
                    model_activation_dict,
                    self.model.hook_layer,
                    feature_indices,
                    max_value_indices,
                )
                for feature_idx, feature_data in batch_dfa_results.items():
                    for prompt_idx in range(feature_data.shape[0]):
                        global_prompt_idx = total_prompts + prompt_idx
                        all_dfa_results[feature_idx][global_prompt_idx] = {
                            "dfaValues": feature_data[prompt_idx][
                                "dfa_values"
                            ].tolist(),
                            "dfaTargetIndex": int(
                                feature_data[prompt_idx]["dfa_target_index"]
                            ),
                            "dfaMaxValue": float(
                                feature_data[prompt_idx]["dfa_max_value"]
                            ),
                        }

                total_prompts += len(minibatch)

            # Update the 1st progress bar (fwd passes & getting sequence data dominates the runtime of these computations)
            if progress is not None:
                progress[0].update(1)

        all_feat_acts = torch.cat(all_feat_acts, dim=0)

        return (
            all_feat_acts,
            torch.tensor([]),  # all_resid_post, no longer used
            feature_resid_dir,
            feature_out_dir,
            corrcoef_neurons,
            corrcoef_encoder,
            all_dfa_results,
        )

    @torch.inference_mode()
    def get_model_acts(
        self,
        minibatch_index: int,
        minibatch_tokens: torch.Tensor,
        use_cache: bool = True,
    ) -> Dict[str, torch.Tensor]:
        """
        A function that gets the model activations for a given minibatch of tokens.
        Uses np.memmap for efficient caching.
        """
        if self.cfg.cache_dir is not None:
            cache_path = self.cfg.cache_dir / f"model_activations_{minibatch_index}.pt"
            if use_cache and cache_path.exists():
                activation_dict = load_tensor_dict_torch(cache_path, self.cfg.device)
            else:
                activation_dict = self.model.forward(
                    minibatch_tokens.to("cpu"), return_logits=False
                )
                save_tensor_dict_torch(activation_dict, cache_path)
        else:
            activation_dict = self.model.forward(
                minibatch_tokens.to("cpu"), return_logits=False
            )

        return activation_dict

    @torch.inference_mode()
    def update_rolling_coefficients(
        self,
        model_acts: Float[Tensor, "batch seq d_in"],
        feature_acts: Float[Tensor, "batch seq feats"],
        corrcoef_neurons: RollingCorrCoef | None,
        corrcoef_encoder: RollingCorrCoef | None,
    ) -> None:
        """

        Args:
            model_acts: Float[Tensor, "batch seq d_in"]
                The activations of the model, which the SAE was trained on.
            feature_idx: list[int]
                The features we're computing the activations for. This will be used to index the encoder's weights.
            corrcoef_neurons: Optional[RollingCorrCoef]
                The object storing the minimal data necessary to compute corrcoef between feature activations & neurons.
            corrcoef_encoder: Optional[RollingCorrCoef]
                The object storing the minimal data necessary to compute corrcoef between pairwise feature activations.
        """
        # Update the CorrCoef object between feature activation & neurons
        if corrcoef_neurons is not None:
            corrcoef_neurons.update(
                einops.rearrange(feature_acts, "batch seq feats -> feats (batch seq)"),
                einops.rearrange(model_acts, "batch seq d_in -> d_in (batch seq)"),
            )

        # Update the CorrCoef object between pairwise feature activations
        if corrcoef_encoder is not None:
            corrcoef_encoder.update(
                einops.rearrange(feature_acts, "batch seq feats -> feats (batch seq)"),
                einops.rearrange(feature_acts, "batch seq feats -> feats (batch seq)"),
            )


def save_tensor_dict_torch(tensor_dict: Dict[str, torch.Tensor], filename: Path):
    torch.save(tensor_dict, filename)


def load_tensor_dict_torch(filename: Path, device: str) -> Dict[str, torch.Tensor]:
    return torch.load(
        filename, map_location=torch.device(device)
    )  # Directly load to GPU


class FeatureMaskingContext:
    def __init__(self, sae: SAE, feature_idxs: List[int]):
        self.sae = sae
        self.feature_idxs = feature_idxs
        self.original_weight = {}

    def __enter__(self):

        ## W_dec
        self.original_weight["W_dec"] = getattr(self.sae, "W_dec").data.clone()
        # mask the weight
        masked_weight = self.sae.W_dec[self.feature_idxs]
        # set the weight
        setattr(self.sae, "W_dec", nn.Parameter(masked_weight))

        ## W_enc
        # clone the weight.
        self.original_weight["W_enc"] = getattr(self.sae, "W_enc").data.clone()
        # mask the weight
        masked_weight = self.sae.W_enc[:, self.feature_idxs]
        # set the weight
        setattr(self.sae, "W_enc", nn.Parameter(masked_weight))

        if self.sae.cfg.architecture == "standard":

            ## b_enc
            self.original_weight["b_enc"] = getattr(self.sae, "b_enc").data.clone()
            # mask the weight
            masked_weight = self.sae.b_enc[self.feature_idxs]
            # set the weight
            setattr(self.sae, "b_enc", nn.Parameter(masked_weight))

        elif self.sae.cfg.architecture == "jumprelu":

            ## b_enc
            self.original_weight["b_enc"] = getattr(self.sae, "b_enc").data.clone()
            # mask the weight
            masked_weight = self.sae.b_enc[self.feature_idxs]
            # set the weight
            setattr(self.sae, "b_enc", nn.Parameter(masked_weight))

            ## threshold
            self.original_weight["threshold"] = getattr(
                self.sae, "threshold"
            ).data.clone()
            # mask the weight
            masked_weight = self.sae.threshold[self.feature_idxs]
            # set the weight
            setattr(self.sae, "threshold", nn.Parameter(masked_weight))

        elif self.sae.cfg.architecture == "gated":

            ## b_gate
            self.original_weight["b_gate"] = getattr(self.sae, "b_gate").data.clone()
            # mask the weight
            masked_weight = self.sae.b_gate[self.feature_idxs]
            # set the weight
            setattr(self.sae, "b_gate", nn.Parameter(masked_weight))

            ## r_mag
            self.original_weight["r_mag"] = getattr(self.sae, "r_mag").data.clone()
            # mask the weight
            masked_weight = self.sae.r_mag[self.feature_idxs]
            # set the weight
            setattr(self.sae, "r_mag", nn.Parameter(masked_weight))

            ## b_mag
            self.original_weight["b_mag"] = getattr(self.sae, "b_mag").data.clone()
            # mask the weight
            masked_weight = self.sae.b_mag[self.feature_idxs]
            # set the weight
            setattr(self.sae, "b_mag", nn.Parameter(masked_weight))
        else:
            raise (ValueError("Invalid architecture"))

        return self

    def __exit__(self, exc_type, exc_value, traceback):  # type: ignore

        # set everything back to normal
        for key, value in self.original_weight.items():
            setattr(self.sae, key, nn.Parameter(value))



from copy import deepcopy
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Callable, List

import numpy as np
from dataclasses_json import dataclass_json

from sae_dashboard.components_config import (
    ActsHistogramConfig,
    FeatureTablesConfig,
    LogitsHistogramConfig,
    LogitsTableConfig,
    PromptConfig,
    SequencesConfig,
)
from sae_dashboard.html_fns import HTML, bgColorMap, uColorMap
from sae_dashboard.utils_fns import (
    HistogramData,
    max_or_1,
    to_str_tokens,
    unprocess_str_tok,
)

PRECISION = 4


@dataclass
class DecoderWeightsDistribution:
    n_heads: int
    allocation_by_head: List[float]


@dataclass_json
@dataclass
class FeatureTablesData:
    """
    This contains all the data necessary to make the left-hand tables in prompt-centric visualization. See diagram
    in readme:

        https://github.com/callummcdougall/sae_vis#data_storing_fnspy

    Inputs:
        neuron_alignment...
            The data for the neuron alignment table (each of its 3 columns). In other words, the data containing which
            neurons in the transformer the encoder feature is most aligned with.

        correlated_neurons...
            The data for the correlated neurons table (each of its 3 columns). In other words, the data containing which
            neurons in the transformer are most correlated with the encoder feature.

        correlated_features...
            The data for the correlated features table (each of its 3 columns). In other words, the data containing
            which features in this encoder are most correlated with each other.

        correlated_b_features...
            The data for the correlated features table (each of its 3 columns). In other words, the data containing
            which features in encoder-B are most correlated with those in the original encoder. Note, this one might be
            absent if we're not using a B-encoder.
    """

    neuron_alignment_indices: list[int] = field(default_factory=list)
    neuron_alignment_values: list[float] = field(default_factory=list)
    neuron_alignment_l1: list[float] = field(default_factory=list)
    correlated_neurons_indices: list[int] = field(default_factory=list)
    correlated_neurons_pearson: list[float] = field(default_factory=list)
    correlated_neurons_cossim: list[float] = field(default_factory=list)
    correlated_features_indices: list[int] = field(default_factory=list)
    correlated_features_pearson: list[float] = field(default_factory=list)
    correlated_features_cossim: list[float] = field(default_factory=list)
    correlated_b_features_indices: list[int] = field(default_factory=list)
    correlated_b_features_pearson: list[float] = field(default_factory=list)
    correlated_b_features_cossim: list[float] = field(default_factory=list)

    def _get_html_data(
        self,
        cfg: FeatureTablesConfig,
        decode_fn: Callable[[int | list[int]], str | list[str]],
        id_suffix: str,
        column: int | tuple[int, int],
        component_specific_kwargs: dict[str, Any] = {},
    ) -> HTML:
        """
        Returns the HTML for the left-hand tables, wrapped in a 'grid-column' div.

        Note, we only ever use this obj in the context of the left-hand column of the feature-centric vis, and it's
        always the same width & height, which is why there's no customization available for this function.
        """
        # Read HTML from file, and replace placeholders with real ID values
        html_str = (
            Path(__file__).parent / "html" / "feature_tables_template.html"
        ).read_text()
        html_str = html_str.replace("FEATURE_TABLES_ID", f"feature-tables-{id_suffix}")

        # Create dictionary storing the data
        data: dict[str, list[dict[str, str | float]]] = {}

        # Store the neuron alignment data, if it exists
        if len(self.neuron_alignment_indices) > 0:
            assert len(self.neuron_alignment_indices) >= cfg.n_rows, "Not enough rows!"
            data["neuronAlignment"] = [
                {
                    "index": index,
                    "value": f"{value:+.3f}",
                    "percentageL1": f"{percent_l1:.1%}",
                }
                for index, value, percent_l1 in zip(
                    self.neuron_alignment_indices,
                    self.neuron_alignment_values,
                    self.neuron_alignment_l1,
                )
            ]

        # Store the other 3, if they exist (they're all in the same format, so we can do it in a for loop)
        for name, js_name in zip(
            ["correlated_neurons", "correlated_features", "correlated_b_features"],
            ["correlatedNeurons", "correlatedFeatures", "correlatedFeaturesB"],
        ):
            if len(getattr(self, f"{name}_indices")) > 0:
                # assert len(getattr(self, f"{name}_indices")) >= cfg.n_rows, "Not enough rows!"
                data[js_name] = [
                    {
                        "index": index,
                        "value": f"{value:+.3f}",
                        "percentageL1": f"{percent_L1:+.3f}",
                    }
                    for index, value, percent_L1 in zip(
                        getattr(self, f"{name}_indices")[: cfg.n_rows],
                        getattr(self, f"{name}_pearson")[: cfg.n_rows],
                        getattr(self, f"{name}_cossim")[: cfg.n_rows],
                    )
                ]

        return HTML(
            html_data={column: html_str},
            js_data={"featureTablesData": {id_suffix: data}},
        )


@dataclass_json
@dataclass
class ActsHistogramData(HistogramData):
    def _get_html_data(
        self,
        cfg: ActsHistogramConfig,
        decode_fn: Callable[[int | list[int]], str | list[str]],
        id_suffix: str,
        column: int | tuple[int, int],
        component_specific_kwargs: dict[str, Any] = {},
    ) -> HTML:
        """
        Converts data -> HTML object, for the feature activations histogram (i.e. the histogram over all sampled tokens,
        showing the distribution of activations for this feature).
        """
        # We can't post-hoc change the number of bins, so check this wasn't changed in the config
        # assert cfg.n_bins == len(self.bar_heights),\
        #     "Can't post-hoc change `n_bins` in histogram config - you need to regenerate data."

        # Read HTML from file, and replace placeholders with real ID values
        html_str = (
            Path(__file__).parent / "html" / "acts_histogram_template.html"
        ).read_text()
        html_str = html_str.replace("HISTOGRAM_ACTS_ID", f"histogram-acts-{id_suffix}")

        # Process colors for frequency histogram; it's darker at higher values
        bar_values_normed = [
            (0.4 * max(self.bar_values) + 0.6 * v)
            / max(max(self.bar_values), 1e-6)  # avoid divide by zero
            for v in self.bar_values
        ]
        bar_colors = [bgColorMap(v) for v in bar_values_normed]

        # Next we create the data dict
        data: dict[str, Any] = {
            "y": self.bar_heights,
            "x": self.bar_values,
            "ticks": self.tick_vals,
            "colors": bar_colors,
        }
        if self.title is not None:
            data["title"] = self.title

        return HTML(
            html_data={column: html_str},
            js_data={"actsHistogramData": {id_suffix: data}},
        )


@dataclass_json
@dataclass
class LogitsHistogramData(HistogramData):
    def _get_html_data(
        self,
        cfg: LogitsHistogramConfig,
        decode_fn: Callable[[int | list[int]], str | list[str]],
        id_suffix: str,
        column: int | tuple[int, int],
        component_specific_kwargs: dict[str, Any] = {},
    ) -> HTML:
        """
        Converts data -> HTML object, for the logits histogram (i.e. the histogram over all tokens in the vocab, showing
        the distribution of direct logit effect on that token).
        """
        # We can't post-hoc change the number of bins, so check this wasn't changed in the config
        # assert cfg.n_bins == len(self.bar_heights),\
        #     "Can't post-hoc change `n_bins` in histogram config - you need to regenerate data."

        # Read HTML from file, and replace placeholders with real ID values
        html_str = (
            Path(__file__).parent / "html" / "logits_histogram_template.html"
        ).read_text()
        html_str = html_str.replace(
            "HISTOGRAM_LOGITS_ID", f"histogram-logits-{id_suffix}"
        )

        data: dict[str, Any] = {
            "y": self.bar_heights,
            "x": self.bar_values,
            "ticks": self.tick_vals,
        }
        if self.title is not None:
            data["title"] = self.title

        return HTML(
            html_data={column: html_str},
            js_data={"logitsHistogramData": {id_suffix: data}},
        )


@dataclass_json
@dataclass
class LogitsTableData:
    bottom_token_ids: list[int] = field(default_factory=list)
    bottom_logits: list[float] = field(default_factory=list)
    top_token_ids: list[int] = field(default_factory=list)
    top_logits: list[float] = field(default_factory=list)

    def _get_html_data(
        self,
        cfg: LogitsTableConfig,
        decode_fn: Callable[[int | list[int]], str | list[str]],
        id_suffix: str,
        column: int | tuple[int, int],
        component_specific_kwargs: dict[str, Any] = {},
    ) -> HTML:
        """
        Converts data -> HTML object, for the logits table (i.e. the top and bottom affected tokens by this feature).
        """
        # Crop the lists to `cfg.n_rows` (first checking the config doesn't ask for more rows than we have)
        assert cfg.n_rows <= len(self.bottom_logits)
        bottom_token_ids = self.bottom_token_ids[: cfg.n_rows]
        bottom_logits = self.bottom_logits[: cfg.n_rows]
        top_token_ids = self.top_token_ids[: cfg.n_rows]
        top_logits = self.top_logits[: cfg.n_rows]

        # Get the negative and positive background values (darkest when equals max abs)
        max_value = max(
            max(top_logits[: cfg.n_rows]), -min(bottom_logits[: cfg.n_rows])
        )
        neg_bg_values = np.absolute(bottom_logits[: cfg.n_rows]) / max_value
        pos_bg_values = np.absolute(top_logits[: cfg.n_rows]) / max_value

        # Get the string tokens, using the decode function
        neg_str = to_str_tokens(decode_fn, bottom_token_ids[: cfg.n_rows])
        pos_str = to_str_tokens(decode_fn, top_token_ids[: cfg.n_rows])

        # Read HTML from file, and replace placeholders with real ID values
        html_str = (
            Path(__file__).parent / "html" / "logits_table_template.html"
        ).read_text()
        html_str = html_str.replace("LOGITS_TABLE_ID", f"logits-table-{id_suffix}")

        # Create object for storing JS data
        data: dict[str, list[dict[str, str | float]]] = {
            "negLogits": [],
            "posLogits": [],
        }

        # Get data for the tables of pos/neg logits
        for i in range(len(neg_str)):
            data["negLogits"].append(
                {
                    "symbol": unprocess_str_tok(neg_str[i]),
                    "value": round(bottom_logits[i], 2),
                    "color": f"rgba(255,{int(255*(1-neg_bg_values[i]))},{int(255*(1-neg_bg_values[i]))},0.5)",
                }
            )
            data["posLogits"].append(
                {
                    "symbol": unprocess_str_tok(pos_str[i]),
                    "value": round(top_logits[i], 2),
                    "color": f"rgba({int(255*(1-pos_bg_values[i]))},{int(255*(1-pos_bg_values[i]))},255,0.5)",
                }
            )

        return HTML(
            html_data={column: html_str},
            js_data={"logitsTableData": {id_suffix: data}},
        )


@dataclass_json
@dataclass
class SequenceData:
    """
    This contains all the data necessary to make a sequence of tokens in the vis. See diagram in readme:

        https://github.com/callummcdougall/sae_vis#data_storing_fnspy

    Always-visible data:
        token_ids:              List of token IDs in the sequence
        feat_acts:              Sizes of activations on this sequence
        loss_contribution:   Effect on loss of this feature, for this particular token (neg = helpful)

    Data which is visible on hover:
        token_logits:       The logits of the particular token in that sequence (used for line on logits histogram)
        top_token_ids:     List of the top 5 logit-boosted tokens by this feature
        top_logits:        List of the corresponding 5 changes in logits for those tokens
        bottom_token_ids:  List of the bottom 5 logit-boosted tokens by this feature
        bottom_logits:     List of the corresponding 5 changes in logits for those tokens
    """

    original_index: int = 0
    qualifying_token_index: int = 0
    token_ids: list[int] = field(default_factory=list)
    feat_acts: list[float] = field(default_factory=list)
    loss_contribution: list[float] = field(default_factory=list)

    token_logits: list[float] = field(default_factory=list)
    top_token_ids: list[list[int]] = field(default_factory=list)
    top_logits: list[list[float]] = field(default_factory=list)
    bottom_token_ids: list[list[int]] = field(default_factory=list)
    bottom_logits: list[list[float]] = field(default_factory=list)

    def __post_init__(self) -> None:
        """
        Filters the logits & token IDs by removing any elements which are zero (this saves space in the eventual
        JavaScript).
        """
        self.seq_len = len(self.token_ids)
        self.top_logits, self.top_token_ids = self._filter(
            self.top_logits, self.top_token_ids
        )
        self.bottom_logits, self.bottom_token_ids = self._filter(
            self.bottom_logits, self.bottom_token_ids
        )

    def _filter(
        self, float_list: list[list[float]], int_list: list[list[int]]
    ) -> tuple[list[list[float]], list[list[int]]]:
        """
        Filters the list of floats and ints, by removing any elements which are zero. Note - the absolute values of the
        floats are monotonic non-increasing, so we can assume that all the elements we keep will be the first elements
        of their respective lists. Also reduces precisions of feature activations & logits.
        """
        # Next, filter out zero-elements and reduce precision
        float_list = [
            [round(f, PRECISION) for f in floats if abs(f) > 1e-6]
            for floats in float_list
        ]
        int_list = [ints[: len(floats)] for ints, floats in zip(int_list, float_list)]
        return float_list, int_list

    def _get_html_data(
        self,
        cfg: PromptConfig | SequencesConfig,
        decode_fn: Callable[[int | list[int]], str | list[str]],
        id_suffix: str,
        column: int | tuple[int, int],
        component_specific_kwargs: dict[str, Any] = {},
    ) -> HTML:
        """
        Args:

        Returns:
            js_data: list[dict[str, Any]]
                The data for this sequence, in the form of a list of dicts for each token (where the dict stores things
                like token, feature activations, etc).
        """
        assert isinstance(
            cfg, (PromptConfig, SequencesConfig)
        ), f"Invalid config type: {type(cfg)}"
        seq_group_id = component_specific_kwargs.get("seq_group_id", None)
        max_feat_act = component_specific_kwargs.get("max_feat_act", None)
        max_loss_contribution = component_specific_kwargs.get(
            "max_loss_contribution", None
        )
        bold_idx = component_specific_kwargs.get("bold_idx", None)
        permanent_line = component_specific_kwargs.get("permanent_line", False)
        first_in_group = component_specific_kwargs.get("first_in_group", True)
        title = component_specific_kwargs.get("title", None)
        hover_above = component_specific_kwargs.get("hover_above", False)

        # If we didn't supply a sequence group ID, then we assume this sequence is on its own, and give it a unique ID
        if seq_group_id is None:
            seq_group_id = f"prompt-{column:03d}"

        # If we didn't specify bold_idx, then set it to be the midpoint
        if bold_idx is None:
            bold_idx = self.seq_len // 2

        # If we only have data for the bold token, we pad out everything with zeros or empty lists
        only_bold = isinstance(cfg, SequencesConfig) and not (cfg.compute_buffer)
        if only_bold:
            assert bold_idx != "max", "Don't know how to deal with this case yet."
            feat_acts = [
                self.feat_acts[0] if (i == bold_idx) else 0.0
                for i in range(self.seq_len)
            ]
            loss_contribution = [
                self.loss_contribution[0] if (i == bold_idx) + 1 else 0.0
                for i in range(self.seq_len)
            ]
            pos_ids = [
                self.top_token_ids[0] if (i == bold_idx) + 1 else []
                for i in range(self.seq_len)
            ]
            neg_ids = [
                self.bottom_token_ids[0] if (i == bold_idx) + 1 else []
                for i in range(self.seq_len)
            ]
            pos_val = [
                self.top_logits[0] if (i == bold_idx) + 1 else []
                for i in range(self.seq_len)
            ]
            neg_val = [
                self.bottom_logits[0] if (i == bold_idx) + 1 else []
                for i in range(self.seq_len)
            ]
        else:
            feat_acts = deepcopy(self.feat_acts)
            loss_contribution = deepcopy(self.loss_contribution)
            pos_ids = deepcopy(self.top_token_ids)
            neg_ids = deepcopy(self.bottom_token_ids)
            pos_val = deepcopy(self.top_logits)
            neg_val = deepcopy(self.bottom_logits)

        # EXPERIMENT: let's just hardcode everything except feature acts to be 0's for now.
        loss_contribution = [0.0 for _ in range(self.seq_len)]
        pos_ids = [[] for _ in range(self.seq_len)]
        neg_ids = [[] for _ in range(self.seq_len)]
        pos_val = [[] for _ in range(self.seq_len)]
        neg_val = [[] for _ in range(self.seq_len)]
        ### END EXPERIMENT

        # Get values for converting into colors later
        bg_denom = max_feat_act or max_or_1(self.feat_acts)
        u_denom = max_loss_contribution or max_or_1(self.loss_contribution, abs=True)
        bg_values = (np.maximum(feat_acts, 0.0) / max(1e-4, bg_denom)).tolist()
        u_values = (np.array(loss_contribution) / max(1e-4, u_denom)).tolist()

        # If we sent in a prompt rather than this being sliced from a longer sequence, then the pos_ids etc will be shorter
        # than the token list by 1, so we need to pad it at the first token
        if isinstance(cfg, PromptConfig):
            assert (
                len(pos_ids)
                == len(neg_ids)
                == len(pos_val)
                == len(neg_val)
                == len(self.token_ids) - 1
            ), "If this is a single prompt, these lists must be the same length as token_ids or 1 less"
            pos_ids = [[]] + pos_ids
            neg_ids = [[]] + neg_ids
            pos_val = [[]] + pos_val
            neg_val = [[]] + neg_val

        assert (
            len(pos_ids)
            == len(neg_ids)
            == len(pos_val)
            == len(neg_val)
            == len(self.token_ids)
        ), "If this is part of a sequence group etc are given, they must be the same length as token_ids"

        # Process the tokens to get str toks
        toks = to_str_tokens(decode_fn, self.token_ids)
        pos_toks = [to_str_tokens(decode_fn, pos) for pos in pos_ids]
        neg_toks = [to_str_tokens(decode_fn, neg) for neg in neg_ids]

        # Define the JavaScript object which will be used to populate the HTML string
        js_data_list = []

        for i in range(len(self.token_ids)):
            # We might store a bunch of different case-specific data in the JavaScript object for each token. This is
            # done in the form of a disjoint union over different dictionaries (which can each be empty or not), this
            # minimizes the size of the overall JavaScript object. See function in `tokens_script.js` for more.
            kwargs_bold: dict[str, bool] = {}
            kwargs_hide: dict[str, bool] = {}
            kwargs_this_token_active: dict[str, Any] = {}
            kwargs_prev_token_active: dict[str, Any] = {}
            kwargs_hover_above: dict[str, bool] = {}

            # Get args if this is the bolded token (we make it bold, and maybe add permanent line to histograms)
            if bold_idx is not None:
                kwargs_bold["isBold"] = (bold_idx == i) or (
                    bold_idx == "max" and i == np.argmax(feat_acts).item()
                )
                if kwargs_bold["isBold"] and permanent_line:
                    kwargs_bold["permanentLine"] = True

            # If we only have data for the bold token, we hide all other tokens' hoverdata (and skip other kwargs)
            if (
                only_bold
                and isinstance(bold_idx, int)
                and (i not in {bold_idx, bold_idx + 1})
            ):
                kwargs_hide["hide"] = True

            else:
                # Get args if we're making the tooltip hover above token (default is below)
                if hover_above:
                    kwargs_hover_above["hoverAbove"] = True

                # If feature active on this token, get background color and feature act (for hist line)
                if abs(feat_acts[i]) > 1e-8:
                    kwargs_this_token_active = dict(
                        featAct=round(feat_acts[i], PRECISION),
                        bgColor=bgColorMap(bg_values[i]),
                    )

                # If prev token active, get the top/bottom logits table, underline color, and loss effect (for hist line)
                pos_toks_i, neg_toks_i, pos_val_i, neg_val_i = (
                    pos_toks[i],
                    neg_toks[i],
                    pos_val[i],
                    neg_val[i],
                )
                if len(pos_toks_i) + len(neg_toks_i) > 0:
                    # Create dictionary
                    kwargs_prev_token_active = dict(
                        posToks=pos_toks_i,
                        negToks=neg_toks_i,
                        posVal=pos_val_i,
                        negVal=neg_val_i,
                        lossEffect=round(loss_contribution[i], PRECISION),
                        uColor=uColorMap(u_values[i]),
                    )

            js_data_list.append(
                dict(
                    tok=unprocess_str_tok(toks[i]),
                    tokID=self.token_ids[i],
                    tokenLogit=round(self.token_logits[i], PRECISION),
                    **kwargs_bold,
                    **kwargs_this_token_active,
                    **kwargs_prev_token_active,
                    **kwargs_hover_above,
                )
            )

        # Create HTML string (empty by default since sequences are added by JavaScript) and JS data
        html_str = ""
        js_seq_group_data: dict[str, Any] = {"data": [js_data_list]}

        # Add group-specific stuff if this is the first sequence in the group
        if first_in_group:
            # Read HTML from file, replace placeholders with real ID values
            html_str = (
                Path(__file__).parent / "html" / "sequences_group_template.html"
            ).read_text()
            html_str = html_str.replace("SEQUENCE_GROUP_ID", seq_group_id)

            # Get title of sequence group, and the idSuffix to match up with a histogram
            js_seq_group_data["idSuffix"] = id_suffix
            if title is not None:
                js_seq_group_data["title"] = title

        return HTML(
            html_data={column: html_str},
            js_data={"tokenData": {seq_group_id: js_seq_group_data}},
        )


@dataclass_json
@dataclass
class SequenceGroupData:
    """
    This contains all the data necessary to make a single group of sequences (e.g. a quantile in prompt-centric
    visualization). See diagram in readme:

        https://github.com/callummcdougall/sae_vis#data_storing_fnspy

    Inputs:
        title:      The title that this sequence group will have, if any. This is used in `_get_html_data`. The titles
                    will actually be in the HTML strings, not in the JavaScript data.
        seq_data:   The data for the sequences in this group.
    """

    title: str = ""
    seq_data: list[SequenceData] = field(default_factory=list)

    def __len__(self) -> int:
        return len(self.seq_data)

    @property
    def max_feat_act(self) -> float:
        """Returns maximum value of feature activation over all sequences in this group."""
        return max_or_1([act for seq in self.seq_data for act in seq.feat_acts])

    @property
    def max_loss_contribution(self) -> float:
        """Returns maximum value of loss contribution over all sequences in this group."""
        return max_or_1(
            [loss for seq in self.seq_data for loss in seq.loss_contribution], abs=True
        )

    def _get_html_data(
        self,
        cfg: SequencesConfig,
        decode_fn: Callable[[int | list[int]], str | list[str]],
        id_suffix: str,
        column: int | tuple[int, int],
        component_specific_kwargs: dict[str, Any] = {},
        # These default values should be correct when we only have one sequence group, because when we call this from
        # a SequenceMultiGroupData we'll override them)
    ) -> HTML:
        """
        This creates a single group of sequences, i.e. title plus some number of vertically stacked sequences.

        Note, `column` is treated specially here, because the col might overflow (hence colulmn could be a tuple).

        Args (from component-specific kwargs):
            seq_group_id:   The id of the sequence group div. This will usually be passed as e.g. "seq-group-001".
            group_size:     Max size of sequences in the group (i.e. we truncate after this many, if argument supplied).
            max_feat_act:   If supplied, then we use this as the most extreme value (for coloring by feature act).

        Returns:
            html_obj:       Object containing the HTML and JavaScript data for this seq group.
        """
        seq_group_id = component_specific_kwargs.get("seq_group_id", None)
        group_size = component_specific_kwargs.get("group_size", None)
        max_feat_act = component_specific_kwargs.get("max_feat_act", self.max_feat_act)
        max_loss_contribution = component_specific_kwargs.get(
            "max_loss_contribution", self.max_loss_contribution
        )

        # Get the data that will go into the div (list of list of dicts, i.e. containing all data for seqs in group). We
        # start with the title.
        html_obj = HTML()

        # If seq_group_id is not supplied, then we assume this is the only sequence in the column, and we name the group
        # after the column
        if seq_group_id is None:
            seq_group_id = f"seq-group-{column:03d}"

        # Accumulate the HTML data for each sequence in this group
        for i, seq in enumerate(self.seq_data[:group_size]):
            html_obj += seq._get_html_data(
                cfg=cfg,
                # pass in a PromptConfig object
                decode_fn=decode_fn,
                id_suffix=id_suffix,
                column=column,
                component_specific_kwargs=dict(
                    bold_idx="max" if cfg.buffer is None else cfg.buffer[0],
                    permanent_line=False,  # in a group, we're never showing a permanent line (only for single seqs)
                    max_feat_act=max_feat_act,
                    max_loss_contribution=max_loss_contribution,
                    seq_group_id=seq_group_id,
                    first_in_group=(i == 0),
                    title=self.title,
                ),
            )

        return html_obj


@dataclass_json
@dataclass
class SequenceMultiGroupData:
    """
    This contains all the data necessary to make multiple groups of sequences (e.g. the different quantiles in the
    prompt-centric visualization). See diagram in readme:

        https://github.com/callummcdougall/sae_vis#data_storing_fnspy
    """

    seq_group_data: list[SequenceGroupData] = field(default_factory=list)

    def __getitem__(self, idx: int) -> SequenceGroupData:
        return self.seq_group_data[idx]

    @property
    def max_feat_act(self) -> float:
        """Returns maximum value of feature activation over all sequences in this group."""
        return max_or_1([seq_group.max_feat_act for seq_group in self.seq_group_data])

    @property
    def max_loss_contribution(self) -> float:
        """Returns maximum value of loss contribution over all sequences in this group."""
        return max_or_1(
            [seq_group.max_loss_contribution for seq_group in self.seq_group_data]
        )

    def _get_html_data(
        self,
        cfg: SequencesConfig,
        decode_fn: Callable[[int | list[int]], str | list[str]],
        id_suffix: str,
        column: int | tuple[int, int],
        component_specific_kwargs: dict[str, Any] = {},
    ) -> HTML:
        """
        Args:
            decode_fn:                  Mapping from token IDs to string tokens.
            id_suffix:                  The suffix for the ID of the div containing the sequences.
            column:                     The index of this column. Note that this will be an int, but we might end up
                                        turning it into a tuple if we overflow into a new column.
            component_specific_kwargs:  Contains any specific kwargs that could be used to customize this component.

        Returns:
            html_obj:  Object containing the HTML and JavaScript data for these multiple seq groups.
        """
        assert isinstance(column, int)

        # Get max activation value & max loss contributions, over all sequences in all groups
        max_feat_act = component_specific_kwargs.get("max_feat_act", self.max_feat_act)
        max_loss_contribution = component_specific_kwargs.get(
            "max_loss_contribution", self.max_loss_contribution
        )

        # Get the correct column indices for the sequence groups, depending on how group_wrap is configured. Note, we
        # deal with overflowing columns by extending the dictionary, i.e. our column argument isn't just `column`, but
        # is a tuple of `(column, x)` where `x` is the number of times we've overflowed. For instance, if we have mode
        # 'stack-none' then our columns are `(column, 0), (column, 1), (column, 1), (column, 1), (column, 2), ...`
        n_groups = len(self.seq_group_data)
        n_quantile_groups = n_groups - 1
        match cfg.stack_mode:
            case "stack-all":
                # Here, we stack all groups into 1st column
                cols = [column for _ in range(n_groups)]
            case "stack-quantiles":
                # Here, we give 1st group its own column, and stack all groups into second column
                cols = [(column, 0)] + [(column, 1) for _ in range(n_quantile_groups)]
            case "stack-none":
                # Here, we stack groups into columns as [1, 3, 3, ...]
                cols = [
                    (column, 0),
                    *[(column, 1 + int(i / 3)) for i in range(n_quantile_groups)],
                ]
            case _:
                raise ValueError(
                    f"Invalid stack_mode: {cfg.stack_mode}. Expected in 'stack-{{all,quantiles,none}}'."
                )

        # Create the HTML object, and add all the sequence groups to it, possibly across different columns
        html_obj = HTML()
        for i, (col, group_size, sequences_group) in enumerate(
            zip(cols, cfg.group_sizes, self.seq_group_data)
        ):
            html_obj += sequences_group._get_html_data(
                cfg=cfg,
                decode_fn=decode_fn,
                id_suffix=id_suffix,
                column=col,
                component_specific_kwargs=dict(
                    group_size=group_size,
                    max_feat_act=max_feat_act,
                    max_loss_contribution=max_loss_contribution,
                    seq_group_id=f"seq-group-{column}-{i}",  # we label our sequence groups with (index, column)
                ),
            )

        return html_obj


GenericData = (
    FeatureTablesData
    | ActsHistogramData
    | LogitsTableData
    | LogitsHistogramData
    | SequenceMultiGroupData
    | SequenceData
)



import json
from typing import Any, Dict, List

import numpy as np
from transformer_lens import HookedTransformer

from sae_dashboard.feature_data import FeatureData
from sae_dashboard.neuronpedia.neuronpedia_dashboard import (
    NeuronpediaDashboardActivation,
    NeuronpediaDashboardBatch,
    NeuronpediaDashboardFeature,
)
from sae_dashboard.neuronpedia.neuronpedia_runner_config import NeuronpediaRunnerConfig
from sae_dashboard.sae_vis_data import SaeVisData


class NpEncoder(json.JSONEncoder):
    def default(self, o: Any):
        if isinstance(o, NeuronpediaDashboardBatch):
            return o.to_dict()
        if isinstance(o, np.integer):
            return int(o)
        if isinstance(o, np.floating):
            return float(o)
        if isinstance(o, np.ndarray):
            return o.tolist()
        return super(NpEncoder, self).default(o)


class FeatureProcessor:
    """
    Class for processing feature data.
    """

    @staticmethod
    def round_list(to_round: List[float]) -> List[float]:
        """Round a list of floats to 3 decimal places."""
        return list(np.round(to_round, 3))

    @staticmethod
    def ensure_list(input_value: Any) -> List[Any]:
        """Ensure the input is a list."""
        return [input_value] if not isinstance(input_value, list) else input_value

    @staticmethod
    def to_str_tokens_safe(
        model: HookedTransformer, vocab_dict: Dict[int, str], tokens: Any
    ) -> Any:
        """Convert tokens to string tokens safely."""
        OUT_OF_RANGE_TOKEN = "<|outofrange|>"
        vocab_max_index = model.cfg.d_vocab - 1

        if isinstance(tokens, int):
            return (
                OUT_OF_RANGE_TOKEN if tokens > vocab_max_index else vocab_dict[tokens]
            )

        if isinstance(tokens, list):
            tokens = np.array(tokens)

        str_tokens = [
            vocab_dict[t] if t <= vocab_max_index else OUT_OF_RANGE_TOKEN
            for t in tokens.flatten()
        ]

        return np.reshape(str_tokens, tokens.shape).tolist()


class NeuronpediaConverter:
    """
    Class for converting SaeVisData to Neuronpedia format.
    """

    @staticmethod
    def convert_to_np_json(
        model: HookedTransformer,
        sae_data: SaeVisData,
        np_cfg: NeuronpediaRunnerConfig,
        vocab_dict: Dict[int, str],
    ) -> str:
        """
        Convert SaeVisData to Neuronpedia JSON format.

        Args:
            sae_data (SaeVisData): The SAE visualization data.
            np_cfg (NeuronpediaRunnerConfig): Configuration for Neuronpedia runner.
            vocab_dict (Dict[int, str]): Dictionary mapping token IDs to strings.

        Returns:
            str: JSON string representation of the feature data.
        """
        features_outputs = NeuronpediaConverter._process_features(
            model, sae_data, np_cfg, vocab_dict
        )
        batch_data = NeuronpediaConverter._create_batch_data(np_cfg, features_outputs)
        return json.dumps(batch_data, cls=NpEncoder)

    @staticmethod
    def _process_features(
        model: HookedTransformer,
        sae_data: SaeVisData,
        np_cfg: NeuronpediaRunnerConfig,
        vocab_dict: Dict[int, str],
    ) -> List[NeuronpediaDashboardFeature]:
        """Process all features and create NeuronpediaDashboardFeature objects."""
        features_outputs = []
        for feature_index, feature_data in sae_data.feature_data_dict.items():
            feature_output = NeuronpediaDashboardFeature()
            feature_output.feature_index = feature_index

            NeuronpediaConverter._process_feature_tables(feature_output, feature_data)
            NeuronpediaConverter._process_feature_logits(
                feature_output, feature_data, model, vocab_dict
            )
            NeuronpediaConverter._process_feature_histograms(
                feature_output, feature_data
            )
            NeuronpediaConverter._process_feature_activations(
                feature_output, feature_data, model, vocab_dict
            )
            NeuronpediaConverter._process_feature_decoder_weight_dist(
                feature_output, feature_data
            )

            feature_output.n_prompts_total = np_cfg.n_prompts_total
            feature_output.n_tokens_in_prompt = np_cfg.n_tokens_in_prompt
            feature_output.dataset = np_cfg.huggingface_dataset_path

            features_outputs.append(feature_output)
        return features_outputs

    @staticmethod
    def _process_feature_tables(
        feature_output: NeuronpediaDashboardFeature, feature_data: FeatureData
    ) -> None:
        """Process feature tables data and update the feature output."""
        if feature_data.feature_tables_data:
            feature_output.neuron_alignment_indices = (
                feature_data.feature_tables_data.neuron_alignment_indices
            )
            feature_output.neuron_alignment_values = FeatureProcessor.round_list(
                feature_data.feature_tables_data.neuron_alignment_values
            )
            feature_output.neuron_alignment_l1 = FeatureProcessor.round_list(
                feature_data.feature_tables_data.neuron_alignment_l1
            )
            feature_output.correlated_neurons_indices = (
                feature_data.feature_tables_data.correlated_neurons_indices
            )
            feature_output.correlated_neurons_l1 = FeatureProcessor.round_list(
                feature_data.feature_tables_data.correlated_neurons_cossim
            )
            feature_output.correlated_neurons_pearson = FeatureProcessor.round_list(
                feature_data.feature_tables_data.correlated_neurons_pearson
            )
            feature_output.correlated_features_indices = (
                feature_data.feature_tables_data.correlated_features_indices
            )
            feature_output.correlated_features_l1 = FeatureProcessor.round_list(
                feature_data.feature_tables_data.correlated_features_cossim
            )
            feature_output.correlated_features_pearson = FeatureProcessor.round_list(
                feature_data.feature_tables_data.correlated_features_pearson
            )

    @staticmethod
    def _process_feature_logits(
        feature_output: NeuronpediaDashboardFeature,
        feature_data: FeatureData,
        model: HookedTransformer,
        vocab_dict: Dict[int, str],
    ) -> None:
        """Process feature logits data and update the feature output."""
        top_logits = FeatureProcessor.round_list(
            feature_data.logits_table_data.top_logits
        )
        bottom_logits = FeatureProcessor.round_list(
            feature_data.logits_table_data.bottom_logits
        )

        feature_output.neg_str = FeatureProcessor.ensure_list(
            FeatureProcessor.to_str_tokens_safe(
                model, vocab_dict, feature_data.logits_table_data.bottom_token_ids
            )
        )
        feature_output.neg_values = bottom_logits
        feature_output.pos_str = FeatureProcessor.ensure_list(
            FeatureProcessor.to_str_tokens_safe(
                model, vocab_dict, feature_data.logits_table_data.top_token_ids
            )
        )
        feature_output.pos_values = top_logits

    @staticmethod
    def _process_feature_histograms(
        feature_output: NeuronpediaDashboardFeature, feature_data: FeatureData
    ) -> None:
        """Process feature histogram data and update the feature output."""
        if feature_data.acts_histogram_data.title:
            feature_output.frac_nonzero = (
                float(
                    feature_data.acts_histogram_data.title.split(" = ")[1].split("%")[0]
                )
                / 100
            )
        else:
            feature_output.frac_nonzero = 0

        freq_hist_data = feature_data.acts_histogram_data
        feature_output.freq_hist_data_bar_values = FeatureProcessor.round_list(
            freq_hist_data.bar_values
        )
        feature_output.freq_hist_data_bar_heights = FeatureProcessor.round_list(
            freq_hist_data.bar_heights
        )

        logits_hist_data = feature_data.logits_histogram_data
        feature_output.logits_hist_data_bar_heights = FeatureProcessor.round_list(
            logits_hist_data.bar_heights
        )
        feature_output.logits_hist_data_bar_values = FeatureProcessor.round_list(
            logits_hist_data.bar_values
        )

    @staticmethod
    def _process_feature_decoder_weight_dist(
        feature_output: NeuronpediaDashboardFeature,
        feature_data: FeatureData,
    ) -> None:
        """Process feature logits data and update the feature output."""
        if feature_data.decoder_weights_data:
            feature_output.decoder_weights_dist = (
                feature_data.decoder_weights_data.allocation_by_head
            )

    @staticmethod
    def _process_feature_activations(
        feature_output: NeuronpediaDashboardFeature,
        feature_data: FeatureData,
        model: HookedTransformer,
        vocab_dict: Dict[int, str],
    ) -> None:
        """Process feature activations data and update the feature output."""
        activations = []
        for sequence_group in feature_data.sequence_data.seq_group_data:
            bin_min, bin_max, bin_contains = (
                NeuronpediaConverter._parse_sequence_group_title(sequence_group.title)
            )

            for sequence in sequence_group.seq_data:
                if (
                    sequence.top_token_ids is not None
                    and sequence.bottom_token_ids is not None
                    and sequence.top_logits is not None
                    and sequence.bottom_logits is not None
                ):
                    activation = NeuronpediaConverter._create_activation(
                        sequence,
                        bin_min,
                        bin_max,
                        bin_contains,
                        feature_data,
                        model,
                        vocab_dict,
                    )
                    activations.append(activation)

        feature_output.activations = activations

    @staticmethod
    def _parse_sequence_group_title(title: str) -> tuple[float, float, float]:
        """Parse the sequence group title to extract bin information."""
        bin_min, bin_max, bin_contains = 0, 0, 0
        if "TOP ACTIVATIONS" in title:
            bin_min, bin_max, bin_contains = -1, 99, -1
            try:
                bin_max = float(title.split(" = ")[-1])
            except ValueError:
                print(f"Error parsing top activations: {title}")
        elif "INTERVAL" in title:
            try:
                split = title.split("<br>")
                first_split = split[0].split(" ")
                bin_min = float(first_split[1])
                bin_max = float(first_split[-1])
                second_split = split[1].split(" ")
                bin_contains = float(second_split[-1].rstrip("%")) / 100
            except ValueError:
                print(f"Error parsing interval: {title}")
        return bin_min, bin_max, bin_contains

    @staticmethod
    def _create_activation(
        sequence: Any,
        bin_min: float,
        bin_max: float,
        bin_contains: float,
        feature_data: FeatureData,
        model: HookedTransformer,
        vocab_dict: Dict[int, str],
    ) -> NeuronpediaDashboardActivation:
        """Create a NeuronpediaDashboardActivation object from sequence data."""
        activation = NeuronpediaDashboardActivation()
        activation.bin_min = bin_min
        activation.bin_max = bin_max
        activation.bin_contains = bin_contains

        if feature_data.dfa_data is not None:
            if sequence.original_index in feature_data.dfa_data:
                dfa_data = feature_data.dfa_data[sequence.original_index]
                # Round DFA values to three decimal points
                activation.dfa_values = [
                    round(v, 3) for v in dfa_data["dfaValues"][1:]
                ]  # Skip BOS token
                activation.dfa_maxValue = round(
                    max(activation.dfa_values), 3
                )  # Recalculate max to skip BOS token
                activation.dfa_targetIndex = (
                    dfa_data["dfaTargetIndex"] - 1
                )  # Adjust for BOS token
            else:
                # print(
                #     f"Warning: DFA data not found for sequence index {sequence.original_index}"
                # )
                activation.dfa_values = []
                activation.dfa_maxValue = 0
                activation.dfa_targetIndex = -1

        activation.tokens = [
            FeatureProcessor.to_str_tokens_safe(model, vocab_dict, token_id)
            for token_id in sequence.token_ids
        ]
        activation.values = FeatureProcessor.round_list(sequence.feat_acts)

        activation.qualifying_token_index = sequence.qualifying_token_index - 1

        return activation

    @staticmethod
    def _create_batch_data(
        np_cfg: NeuronpediaRunnerConfig,
        features_outputs: List[NeuronpediaDashboardFeature],
    ) -> NeuronpediaDashboardBatch:
        """Create a NeuronpediaDashboardBatch object from processed features."""
        batch_data = NeuronpediaDashboardBatch()

        if np_cfg.model_id is not None and np_cfg.layer is not None:
            batch_data.model_id = np_cfg.model_id
            batch_data.layer = np_cfg.layer
        batch_data.sae_set = (
            np_cfg.sae_set if not np_cfg.np_set_name else np_cfg.np_set_name
        )
        if np_cfg.np_sae_id_suffix is not None:
            batch_data.sae_id_suffix = np_cfg.np_sae_id_suffix
        batch_data.features = features_outputs

        return batch_data



from dataclasses import dataclass
from typing import Any, List, Optional

import numpy as np

EQUAL_VALUE_TOLERANCE = 0.1


# function to check that each value in a list is a float
def check_list_floats(li: list[float]):
    for i in range(len(li)):
        if not isinstance(li[i], float):
            return False
    return True


def equalish(a: Any, b: Any, tol: float = EQUAL_VALUE_TOLERANCE):
    assert type(a) == type(b), f"types do not match: {type(a)} and {type(b)}"

    if (
        isinstance(a, list)
        and isinstance(b, list)
        and check_list_floats(a)
        and check_list_floats(b)
    ):
        close = np.allclose(a, b, tol)
        if not close:
            print(f"Does not match within tolerance: {a} and {b} with tolerance {tol}")
        return close
    elif isinstance(a, float) and isinstance(b, float):
        return abs(a - b) < tol
    else:
        return a == b


@dataclass
class NeuronpediaDashboardActivation:

    def __init__(
        self,
        bin_min: float = 0,
        bin_max: float = 0,
        bin_contains: float = 0,
        tokens: list[str] = [],
        values: list[float] = [],
        qualifying_token_index: int = 0,
        dfa_values: Optional[List[float]] = None,
        dfa_maxValue: Optional[float] = None,
        dfa_targetIndex: Optional[int] = None,
    ):
        self.bin_min = bin_min
        self.bin_max = bin_max
        self.bin_contains = bin_contains
        self.tokens = tokens
        self.values = values
        self.qualifying_token_index = qualifying_token_index
        self.dfa_values = dfa_values
        self.dfa_maxValue = dfa_maxValue
        self.dfa_targetIndex = dfa_targetIndex

    def __eq__(self, other: Any):
        if equalish(self.bin_min, other.bin_min) is False:
            print(f"bin_min does not match: {self.bin_min} and {other.bin_min}")
            return False
        if equalish(self.bin_max, other.bin_max) is False:
            print(f"bin_max does not match: {self.bin_max} and {other.bin_max}")
            return False
        if equalish(self.bin_contains, other.bin_contains, 0.001) is False:
            print(
                f"bin_contains does not match: {self.bin_contains} and {other.bin_contains}"
            )
            return False
        if self.tokens != other.tokens:
            print(f"tokens does not match: {self.tokens} and {other.tokens}")
            return False
        if equalish(self.values, other.values, 0.5) is False:
            print(f"values does not match: {self.values} and {other.values}")
            return False
        return True

    def to_dict(self):
        res = {
            "bin_min": self.bin_min,
            "bin_max": self.bin_max,
            "bin_contains": self.bin_contains,
            "tokens": self.tokens,
            "values": self.values,
            "qualifying_token_index": self.qualifying_token_index,
        }
        if self.dfa_values is not None:
            res["dfa_values"] = self.dfa_values
        if self.dfa_maxValue is not None:
            res["dfa_maxValue"] = self.dfa_maxValue
        if self.dfa_targetIndex is not None:
            res["dfa_targetIndex"] = self.dfa_targetIndex

        return res


@dataclass
class NeuronpediaDashboardFeature:

    def __init__(
        self,
        feature_index: int = 0,
        neuron_alignment_indices: list[int] = [],
        neuron_alignment_values: list[float] = [],
        neuron_alignment_l1: list[float] = [],
        correlated_neurons_indices: list[int] = [],
        correlated_neurons_l1: list[float] = [],
        correlated_neurons_pearson: list[float] = [],
        correlated_features_indices: list[int] = [],
        correlated_features_l1: list[float] = [],
        correlated_features_pearson: list[float] = [],
        neg_str: list[str] = [],
        neg_values: list[float] = [],
        pos_str: list[str] = [],
        pos_values: list[float] = [],
        frac_nonzero: float = 0,
        freq_hist_data_bar_values: list[float] = [],
        freq_hist_data_bar_heights: list[float] = [],
        logits_hist_data_bar_heights: list[float] = [],
        logits_hist_data_bar_values: list[float] = [],
        n_prompts_total: int = 0,
        n_tokens_in_prompt: int = 0,
        dataset: str = "",
        activations: list[dict[str, Any]] = [],
        decoder_weights_dist: list[float] = [],
    ):
        self.feature_index = feature_index
        self.neuron_alignment_indices = neuron_alignment_indices
        self.neuron_alignment_values = neuron_alignment_values
        self.neuron_alignment_l1 = neuron_alignment_l1
        self.correlated_neurons_indices = correlated_neurons_indices
        self.correlated_neurons_l1 = correlated_neurons_l1
        self.correlated_neurons_pearson = correlated_neurons_pearson
        self.correlated_features_indices = correlated_features_indices
        self.correlated_features_l1 = correlated_features_l1
        self.correlated_features_pearson = correlated_features_pearson
        self.neg_str = neg_str
        self.neg_values = neg_values
        self.pos_str = pos_str
        self.pos_values = pos_values
        self.frac_nonzero = frac_nonzero
        self.freq_hist_data_bar_values = freq_hist_data_bar_values
        self.freq_hist_data_bar_heights = freq_hist_data_bar_heights
        self.logits_hist_data_bar_heights = logits_hist_data_bar_heights
        self.logits_hist_data_bar_values = logits_hist_data_bar_values
        self.n_prompts_total = n_prompts_total
        self.n_tokens_in_prompt = n_tokens_in_prompt
        self.dataset = dataset
        self.activations: list[NeuronpediaDashboardActivation] = []
        self.decoder_weights_dist = decoder_weights_dist
        for activation in activations:
            self.activations.append(NeuronpediaDashboardActivation(**activation))

    def __eq__(self, other: Any):
        if self.feature_index != other.feature_index:
            print(
                f"feature_index does not match: {self.feature_index} and {other.feature_index}"
            )
            return False
        if self.neuron_alignment_indices != other.neuron_alignment_indices:
            print(
                f"neuron_alignment_indices does not match: {self.neuron_alignment_indices} and {other.neuron_alignment_indices}"
            )
            return False
        if (
            equalish(self.neuron_alignment_values, other.neuron_alignment_values)
            is False
        ):
            print(equalish(self.neuron_alignment_values, other.neuron_alignment_values))
            print(
                f"neuron_alignment_values does not match: {self.neuron_alignment_values} and {other.neuron_alignment_values}"
            )
            return False
        if equalish(self.neuron_alignment_l1, other.neuron_alignment_l1) is False:
            print(
                f"neuron_alignment_l1 does not match: {self.neuron_alignment_l1} and {other.neuron_alignment_l1}"
            )
            return False

        if self.neg_str != other.neg_str:
            print(f"neg_str does not match: {self.neg_str} and {other.neg_str}")
            return False
        if equalish(self.neg_values, other.neg_values) is False:
            print(
                f"neg_values does not match: {self.neg_values} and {other.neg_values}"
            )
            return False
        if self.pos_str != other.pos_str:
            print(f"pos_str does not match: {self.pos_str} and {other.pos_str}")
            return False
        if equalish(self.pos_values, other.pos_values) is False:
            print(
                f"pos_values does not match: {self.pos_values} and {other.pos_values}"
            )
            return False
        if equalish(self.frac_nonzero, other.frac_nonzero, 0.001) is False:
            print(
                f"frac_nonzero does not match: {self.frac_nonzero} and {other.frac_nonzero}"
            )
            return False
        if (
            equalish(self.freq_hist_data_bar_values, other.freq_hist_data_bar_values)
            is False
        ):
            print(
                f"freq_hist_data_bar_values does not match: {self.freq_hist_data_bar_values} and {other.freq_hist_data_bar_values}"
            )
            return False
        if self.freq_hist_data_bar_heights != other.freq_hist_data_bar_heights:
            print(
                f"freq_hist_data_bar_heights does not match: {self.freq_hist_data_bar_heights} and {other.freq_hist_data_bar_heights}"
            )
            return False
        if self.logits_hist_data_bar_heights != other.logits_hist_data_bar_heights:
            print(
                f"logits_hist_data_bar_heights does not match: {self.logits_hist_data_bar_heights} and {other.logits_hist_data_bar_heights}"
            )
            return False
        if (
            equalish(
                self.logits_hist_data_bar_values, other.logits_hist_data_bar_values
            )
            is False
        ):
            print(
                f"logits_hist_data_bar_values does not match: {self.logits_hist_data_bar_values} and {other.logits_hist_data_bar_values}"
            )
            return False
        if self.n_prompts_total != other.n_prompts_total:
            print(
                f"n_prompts_total does not match: {self.n_prompts_total} and {other.n_prompts_total}"
            )
            return False
        if self.n_tokens_in_prompt != other.n_tokens_in_prompt:
            print(
                f"n_tokens_in_prompt does not match: {self.n_tokens_in_prompt} and {other.n_tokens_in_prompt}"
            )
            return False
        if self.dataset != other.dataset:
            print(f"dataset does not match: {self.dataset} and {other.dataset}")
            return False
        for i, activation in enumerate(self.activations):
            if activation != other.activations[i]:
                print("".join(other.activations[i].tokens))
                print(" ==================================== ")
                print("".join(activation.tokens))
                print(
                    f"activation {i} does not match: {activation} and {other.activations[i]}"
                )
                return False
        if self.decoder_weights_dist != other.decoder_weights_dist:
            print(
                f"decoder_weights_dist does not match: {self.decoder_weights_dist} and {other.decoder_weights_dist}"
            )
            return False
        return True

    def to_dict(self):
        return {
            "feature_index": self.feature_index,
            "neuron_alignment_indices": self.neuron_alignment_indices,
            "neuron_alignment_values": self.neuron_alignment_values,
            "neuron_alignment_l1": self.neuron_alignment_l1,
            "correlated_neurons_indices": self.correlated_neurons_indices,
            "correlated_neurons_l1": self.correlated_neurons_l1,
            "correlated_neurons_pearson": self.correlated_neurons_pearson,
            "correlated_features_indices": self.correlated_features_indices,
            "correlated_features_l1": self.correlated_features_l1,
            "correlated_features_pearson": self.correlated_features_pearson,
            "neg_str": self.neg_str,
            "neg_values": self.neg_values,
            "pos_str": self.pos_str,
            "pos_values": self.pos_values,
            "frac_nonzero": self.frac_nonzero,
            "freq_hist_data_bar_values": self.freq_hist_data_bar_values,
            "freq_hist_data_bar_heights": self.freq_hist_data_bar_heights,
            "logits_hist_data_bar_heights": self.logits_hist_data_bar_heights,
            "logits_hist_data_bar_values": self.logits_hist_data_bar_values,
            "n_prompts_total": self.n_prompts_total,
            "n_tokens_in_prompt": self.n_tokens_in_prompt,
            "dataset": self.dataset,
            "decoder_weights_dist": self.decoder_weights_dist,
            "activations": [activation.to_dict() for activation in self.activations],
        }


# TODO: just add the NPRunnerConfig instead


@dataclass
class NeuronpediaDashboardBatch:

    def __init__(
        self,
        model_id: str = "",
        layer: int = 0,
        sae_set: str = "",
        sae_id_suffix: Optional[str] = None,
        features: list[dict[str, Any]] = [],
        # settings: NeuronpediaDashboardSettings = NeuronpediaDashboardSettings(),
    ):
        self.model_id = model_id
        self.layer = layer
        self.sae_set = sae_set
        self.sae_id_suffix = sae_id_suffix
        self.features: list[NeuronpediaDashboardFeature] = []
        for feature in features:
            self.features.append(NeuronpediaDashboardFeature(**feature))
        # self.settings = settings

    def __eq__(self, other: Any):
        if self.model_id != other.model_id:
            print(f"model_id does not match: {self.model_id} and {other.model_id}")
            return False
        if self.layer != other.layer:
            print(f"layer does not match: {self.layer} and {other.layer}")
            return False
        if self.sae_set != other.sae_set:
            print(f"sae_set does not match: {self.sae_set} and {other.sae_set}")
            return False
        if self.sae_id_suffix != other.sae:
            print(
                f"sae_id_suffix does not match: {self.sae_id_suffix} and {other.sae_id_suffix}"
            )
            return False
        for i, feature in enumerate(self.features):
            if feature != other.features[i]:
                print(
                    f"feature {feature.feature_index} does not match: {feature} and {other.features[i]}"
                )
                return False
        # if self.settings != other.settings:
        #     print(f"settings does not match: {self.settings} and {other.settings}")
        #     return False
        return True

    def to_dict(self):
        return {
            "model_id": self.model_id,
            "layer": self.layer,
            "sae_set": self.sae_set,
            "sae_id_suffix": self.sae_id_suffix,
            "features": [feature.to_dict() for feature in self.features],
            # "settings": self.settings.to_dict(),
        }



from dataclasses import dataclass
from typing import Any, Iterator, Literal

SEQUENCES_CONFIG_HELP = dict(
    buffer="How many tokens to add as context to each sequence, on each side. The tokens chosen for the top acts / \
quantile groups can't be outside the buffer range. If None, we use the entire sequence as context.",
    compute_buffer="If False, then we don't compute the loss effect, activations, or any other data for tokens \
other than the bold tokens in our sequences (saving time).",
    n_quantiles="Number of quantile groups for the sequences. If zero, we only show top activations, no quantile \
groups.",
    top_acts_group_size="Number of sequences in the 'top activating sequences' group.",
    quantile_group_size="Number of sequences in each of the sequence quantile groups.",
    top_logits_hoverdata="Number of top/bottom logits to show in the hoverdata for each token.",
    stack_mode="How to stack the sequence groups.\n  'stack-all' = all groups are stacked in a single column \
(scrolls vertically if it overflows)\n  'stack-quantiles' = first col contains top acts, second col contains all \
quantile groups\n  'stack-none' = we stack in a way which ensures no vertical scrolling.",
    hover_below="Whether the hover information about a token appears below or above the token.",
)

ACTIVATIONS_HISTOGRAM_CONFIG_HELP = dict(
    n_bins="Number of bins for the histogram.",
)

LOGITS_HISTOGRAM_CONFIG_HELP = dict(
    n_bins="Number of bins for the histogram.",
)

LOGITS_TABLE_CONFIG_HELP = dict(
    n_rows="Number of top/bottom logits to show in the table.",
)

FEATURE_TABLES_CONFIG_HELP = dict(
    n_rows="Number of rows to show for each feature table.",
    neuron_alignment_table="Whether to show the neuron alignment table.",
    correlated_neurons_table="Whether to show the correlated neurons table.",
    correlated_features_table="Whether to show the (pairwise) correlated features table.",
    correlated_b_features_table="Whether to show the correlated encoder-B features table.",
)


@dataclass
class BaseComponentConfig:
    def data_is_contained_in(self, other: "BaseComponentConfig") -> bool:
        """
        This returns False only when the data that was computed based on `other` wouldn't be enough to show the data
        that was computed based on `self`. For instance, if `self` was a config object with 10 rows, and `other` had
        just 5 rows, then this would return False. A less obvious example: if `self` was a histogram config with 50 bins
        then `other` would need to have exactly 50 bins (because we can't change the bins after generating them).
        """
        return True

    @property
    def help_dict(self) -> dict[str, str]:
        """
        This is a dictionary which maps the name of each argument to a description of what it does. This is used when
        printing out the help for a config object, to show what each argument does.
        """
        return {}


@dataclass
class PromptConfig(BaseComponentConfig):
    pass


@dataclass
class SequencesConfig(BaseComponentConfig):
    buffer: tuple[int, int] | None = (5, 5)
    compute_buffer: bool = True
    n_quantiles: int = 10
    top_acts_group_size: int = 20
    quantile_group_size: int = 5
    top_logits_hoverdata: int = 5
    stack_mode: Literal["stack-all", "stack-quantiles", "stack-none"] = "stack-all"
    hover_below: bool = True

    def data_is_contained_in(self, other: BaseComponentConfig) -> bool:
        assert isinstance(other, self.__class__)
        return all(
            [
                self.buffer is None
                or (
                    other.buffer is not None and self.buffer[0] <= other.buffer[0]
                ),  # the buffer needs to be <=
                self.buffer is None
                or (other.buffer is not None and self.buffer[1] <= other.buffer[1]),
                int(self.compute_buffer)
                <= int(
                    other.compute_buffer
                ),  # we can't compute the buffer if we didn't in `other`
                self.n_quantiles
                in {
                    0,
                    other.n_quantiles,
                },  # we actually need the quantiles identical (or one to be zero)
                self.top_acts_group_size
                <= other.top_acts_group_size,  # group size needs to be <=
                self.quantile_group_size
                <= other.quantile_group_size,  # each quantile group needs to be <=
                self.top_logits_hoverdata
                <= other.top_logits_hoverdata,  # hoverdata rows need to be <=
            ]
        )

    def __post_init__(self):
        # Get list of group lengths, based on the config params
        self.group_sizes = [self.top_acts_group_size] + [
            self.quantile_group_size
        ] * self.n_quantiles

    @property
    def help_dict(self) -> dict[str, str]:
        return SEQUENCES_CONFIG_HELP


@dataclass
class ActsHistogramConfig(BaseComponentConfig):
    n_bins: int = 50

    def data_is_contained_in(self, other: BaseComponentConfig) -> bool:
        assert isinstance(other, self.__class__)
        return self.n_bins == other.n_bins

    @property
    def help_dict(self) -> dict[str, str]:
        return ACTIVATIONS_HISTOGRAM_CONFIG_HELP


@dataclass
class LogitsHistogramConfig(BaseComponentConfig):
    n_bins: int = 50

    def data_is_contained_in(self, other: BaseComponentConfig) -> bool:
        assert isinstance(other, self.__class__)
        return self.n_bins == other.n_bins

    @property
    def help_dict(self) -> dict[str, str]:
        return LOGITS_HISTOGRAM_CONFIG_HELP


@dataclass
class LogitsTableConfig(BaseComponentConfig):
    n_rows: int = 10

    def data_is_contained_in(self, other: BaseComponentConfig) -> bool:
        assert isinstance(other, self.__class__)
        return self.n_rows <= other.n_rows

    @property
    def help_dict(self) -> dict[str, str]:
        return LOGITS_TABLE_CONFIG_HELP


@dataclass
class FeatureTablesConfig(BaseComponentConfig):
    n_rows: int = 3
    neuron_alignment_table: bool = True
    correlated_neurons_table: bool = True
    correlated_features_table: bool = True
    correlated_b_features_table: bool = False

    def data_is_contained_in(self, other: BaseComponentConfig) -> bool:
        assert isinstance(other, self.__class__)
        return all(
            [
                self.n_rows <= other.n_rows,
                self.neuron_alignment_table <= other.neuron_alignment_table,
                self.correlated_neurons_table <= other.correlated_neurons_table,
                self.correlated_features_table <= other.correlated_features_table,
                self.correlated_b_features_table <= other.correlated_b_features_table,
            ]
        )

    @property
    def help_dict(self) -> dict[str, str]:
        return FEATURE_TABLES_CONFIG_HELP


GenericComponentConfig = (
    PromptConfig
    | SequencesConfig
    | ActsHistogramConfig
    | LogitsHistogramConfig
    | LogitsTableConfig
    | FeatureTablesConfig
)


class Column:
    def __init__(
        self,
        *args: GenericComponentConfig,
        width: int | None = None,
    ):
        self.components = list(args)
        self.width = width

    def __iter__(self) -> Iterator[Any]:
        return iter(self.components)

    def __getitem__(self, idx: int) -> Any:
        return self.components[idx]

    def __len__(self) -> int:
        return len(self.components)



import math
import random
import re
from collections import defaultdict
from typing import Iterable, List, Union

import einops
import numpy as np
import torch
from jaxtyping import Int
from rich import print as rprint
from rich.table import Table
from sae_lens import SAE
from sae_lens.config import DTYPE_MAP as DTYPES
from torch import Tensor
from tqdm.auto import tqdm
from transformer_lens import HookedTransformer

from sae_dashboard.components import (
    ActsHistogramData,
    DecoderWeightsDistribution,
    FeatureTablesData,
    LogitsHistogramData,
)
from sae_dashboard.data_parsing_fns import (
    get_features_table_data,
    get_logits_table_data,
)
from sae_dashboard.feature_data import FeatureData
from sae_dashboard.feature_data_generator import FeatureDataGenerator
from sae_dashboard.sae_vis_data import SaeVisConfig, SaeVisData
from sae_dashboard.sequence_data_generator import SequenceDataGenerator
from sae_dashboard.transformer_lens_wrapper import (
    ActivationConfig,
    TransformerLensWrapper,
)
from sae_dashboard.utils_fns import FeatureStatistics


class FeatureDataGeneratorFactory:
    @staticmethod
    def create(
        cfg: SaeVisConfig,
        model: HookedTransformer,
        encoder: SAE,
        tokens: Int[Tensor, "batch seq"],
    ) -> FeatureDataGenerator:
        """Builds a FeatureDataGenerator using the provided config and model."""
        activation_config = ActivationConfig(
            primary_hook_point=cfg.hook_point,
            auxiliary_hook_points=(
                [
                    re.sub(r"hook_z", "hook_v", cfg.hook_point),
                    re.sub(r"hook_z", "hook_pattern", cfg.hook_point),
                ]
                if cfg.use_dfa
                else []
            ),
        )
        wrapped_model = TransformerLensWrapper(model, activation_config)
        return FeatureDataGenerator(
            cfg=cfg, model=wrapped_model, encoder=encoder, tokens=tokens
        )


class SaeVisRunner:
    def __init__(self, cfg: SaeVisConfig) -> None:
        self.cfg = cfg
        self.device = self.cfg.device
        self.dtype = DTYPES[self.cfg.dtype]
        if self.cfg.cache_dir is not None:
            self.cfg.cache_dir.mkdir(parents=True, exist_ok=True)

    @torch.inference_mode()
    def run(
        self, encoder: SAE, model: HookedTransformer, tokens: Int[Tensor, "batch seq"]
    ) -> SaeVisData:
        # Apply random seed
        self.set_seeds()

        # add extra method to SAE which is not yet provided by SAE Lens.
        # encoder = self.mock_feature_acts_subset_for_now(encoder)
        encoder.fold_W_dec_norm()

        # turn off reshaping mode since that's not useful if we're caching activations on disk
        if encoder.hook_z_reshaping_mode:
            encoder.turn_off_forward_pass_hook_z_reshaping()

        # set precision on encoders and model
        # encoder = encoder.to(DTYPES[self.cfg.dtype])
        # # model = cast(HookedTransformer, model.to(DTYPES[self.cfg.dtype]))

        # Create objects to store all the data we'll get from `_get_feature_data`
        sae_vis_data = SaeVisData(cfg=self.cfg)
        # model.to(self.cfg.device)
        # encoder = encoder.to(self.cfg.device)
        time_logs = defaultdict(float)

        features_list = self.handle_features(self.cfg.features, encoder)
        feature_batches = self.get_feature_batches(features_list)
        progress = self.get_progress_bar(tokens, feature_batches, features_list)

        feature_data_generator = FeatureDataGeneratorFactory.create(
            self.cfg, model, encoder, tokens
        )

        sequence_data_generator = SequenceDataGenerator(
            cfg=self.cfg,
            tokens=tokens,
            W_U=model.W_U,
        )

        all_consolidated_dfa_results = {
            feature_idx: {} for feature_idx in self.cfg.features
        }
        # For each batch of features: get new data and update global data storage objects
        # TODO: We should write out json files with the results as this runs rather than storing everything in memory
        for features in feature_batches:
            # model and sae activations calculations.

            (
                all_feat_acts,
                _,  # all resid post. no longer used.
                feature_resid_dir,
                feature_out_dir,
                corrcoef_neurons,
                corrcoef_encoder,
                batch_dfa_results,
            ) = feature_data_generator.get_feature_data(features, progress)

            # Get the logits of all features (i.e. the directions this feature writes to the logit output)
            logits = einops.einsum(
                feature_resid_dir.to(device=model.W_U.device, dtype=model.W_U.dtype),
                model.W_U,
                "feats d_model, d_model d_vocab -> feats d_vocab",
            ).to(self.device)

            # ! Get stats (including quantiles, which will be useful for the prompt-centric visualisation)
            feature_stats = FeatureStatistics.create(
                data=einops.rearrange(all_feat_acts, "b s feats -> feats (b s)"),
                batch_size=self.cfg.quantile_feature_batch_size,
            )

            # ! Data setup code (defining the main objects we'll eventually return)
            feature_data_dict: dict[int, FeatureData] = {
                feat: FeatureData() for feat in features
            }

            # We're using `cfg.feature_centric_layout` to figure out what data we'll need to calculate during this function
            layout = self.cfg.feature_centric_layout

            feature_tables_data = get_features_table_data(
                feature_out_dir=feature_out_dir,
                corrcoef_neurons=corrcoef_neurons,
                corrcoef_encoder=corrcoef_encoder,
                n_rows=layout.feature_tables_cfg.n_rows,  # type: ignore
            )

            # Add all this data to the list of FeatureTablesData objects
            if batch_dfa_results:
                # Accumulate DFA results across feature batches
                for feature_idx, feature_data in batch_dfa_results.items():
                    all_consolidated_dfa_results[feature_idx].update(feature_data)

            for i, (feat, logit_vector) in enumerate(zip(features, logits)):
                feature_data_dict[feat].feature_tables_data = FeatureTablesData(
                    **{k: v[i] for k, v in feature_tables_data.items()}  # type: ignore
                )

                # Get logits histogram data (no title)
                feature_data_dict[feat].logits_histogram_data = (
                    LogitsHistogramData.from_data(
                        data=logit_vector.to(
                            torch.float32
                        ),  # need this otherwise fails on MPS
                        n_bins=layout.logits_hist_cfg.n_bins,  # type: ignore
                        tickmode="5 ticks",
                        title=None,
                    )
                )

                # Get data for feature activations histogram (including the title!)
                feat_acts = all_feat_acts[..., i]

                # Create a mask for tokens to ignore based on both ID and position
                ignore_tokens_mask = torch.ones_like(tokens, dtype=torch.bool)
                if self.cfg.ignore_tokens:
                    ignore_tokens_mask &= ~torch.isin(
                        tokens,
                        torch.tensor(
                            list(self.cfg.ignore_tokens),
                            dtype=tokens.dtype,
                            device=tokens.device,
                        ),
                    )
                if self.cfg.ignore_positions:
                    ignore_positions_mask = torch.ones_like(tokens, dtype=torch.bool)
                    ignore_positions_mask[:, self.cfg.ignore_positions] = False
                    ignore_tokens_mask &= ignore_positions_mask

                # Move the mask to the same device as feat_acts
                ignore_tokens_mask = ignore_tokens_mask.to(feat_acts.device)

                # set any masked positions to 0
                masked_feat_acts = feat_acts * ignore_tokens_mask

                # Apply the mask to feat_acts
                nonzero_feat_acts = masked_feat_acts[masked_feat_acts > 0]
                frac_nonzero = nonzero_feat_acts.numel() / masked_feat_acts.numel()

                feature_data_dict[feat].acts_histogram_data = (
                    ActsHistogramData.from_data(
                        data=nonzero_feat_acts.to(
                            torch.float32
                        ),  # need this otherwise fails on MPS
                        n_bins=layout.act_hist_cfg.n_bins,  # type: ignore
                        tickmode="5 ticks",
                        title=f"ACTIVATIONS<br>DENSITY = {frac_nonzero:.3%}",
                    )
                )

                # Create a MiddlePlotsData object from this, and add it to the dict
                feature_data_dict[feat].logits_table_data = get_logits_table_data(
                    logit_vector=logit_vector,
                    n_rows=layout.logits_table_cfg.n_rows,  # type: ignore
                )

                # ! Calculate all data for the right-hand visualisations, i.e. the sequences

                # Add this feature's sequence data to the list
                feature_data_dict[feat].sequence_data = (
                    sequence_data_generator.get_sequences_data(
                        feat_acts=masked_feat_acts,
                        feat_logits=logits[i],
                        resid_post=torch.tensor([]),  # no longer used
                        feature_resid_dir=feature_resid_dir[i],
                    )
                )
                if self.cfg.use_dfa:
                    feature_data_dict[feat].dfa_data = all_consolidated_dfa_results.get(
                        feat, None
                    )
                    feature_data_dict[feat].decoder_weights_data = (
                        get_decoder_weights_distribution(encoder, model, feat)[0]
                    )

                # Update the 2nd progress bar (fwd passes & getting sequence data dominates the runtime of these computations)
                if progress is not None:
                    progress[1].update(1)

            # ! Return the output, as a dict of FeatureData items
            new_feature_data = SaeVisData(
                cfg=self.cfg,
                feature_data_dict=feature_data_dict,
                feature_stats=feature_stats,
            )

            sae_vis_data.update(new_feature_data)

        # Now exited, make sure the progress bar is at 100%
        if progress is not None:
            for pbar in progress:
                pbar.n = pbar.total

        # If verbose, then print the output
        if self.cfg.verbose:
            total_time = sum(time_logs.values())
            table = Table("Task", "Time", "Pct %")
            for task, duration in time_logs.items():
                table.add_row(task, f"{duration:.2f}s", f"{duration/total_time:.1%}")
            rprint(table)

        sae_vis_data.cfg = self.cfg
        sae_vis_data.model = model
        sae_vis_data.encoder = encoder

        return sae_vis_data

    def set_seeds(self) -> None:
        if self.cfg.seed is not None:
            random.seed(self.cfg.seed)
            torch.manual_seed(self.cfg.seed)
            np.random.seed(self.cfg.seed)
        return None

    def handle_features(
        self, features: Iterable[int] | None, encoder_wrapper: SAE
    ) -> list[int]:
        if features is None:
            return list(range(encoder_wrapper.cfg.d_sae))
        else:
            return list(features)

    def get_feature_batches(self, features_list: list[int]) -> list[list[int]]:
        # Break up the features into batches
        feature_batches = [
            x.tolist()
            for x in torch.tensor(features_list).split(self.cfg.minibatch_size_features)
        ]
        return feature_batches

    def get_progress_bar(
        self,
        tokens: Int[Tensor, "batch seq"],
        feature_batches: list[list[int]],
        features_list: list[int],
    ):
        # Calculate how many minibatches of tokens there will be (for the progress bar)
        n_token_batches = (
            1
            if (self.cfg.minibatch_size_tokens is None)
            else math.ceil(len(tokens) / self.cfg.minibatch_size_tokens)
        )

        # Get the denominator for each of the 2 progress bars
        totals = (n_token_batches * len(feature_batches), len(features_list))

        # Optionally add two progress bars (one for the forward passes, one for getting the sequence data)
        if self.cfg.verbose:
            progress = [
                tqdm(total=totals[0], desc="Forward passes to cache data for vis"),
                tqdm(total=totals[1], desc="Extracting vis data from cached data"),
            ]
        else:
            progress = None

        return progress


def get_decoder_weights_distribution(
    encoder: SAE,
    model: HookedTransformer,
    feature_idx: Union[int, List[int]],
) -> List[DecoderWeightsDistribution]:
    if not isinstance(feature_idx, list):
        feature_idx = [feature_idx]

    distribs = []
    for feature in feature_idx:
        att_blocks = einops.rearrange(
            encoder.W_dec[feature, :],
            "(n_head d_head) -> n_head d_head",
            n_head=model.cfg.n_heads,
        ).to("cpu")
        decoder_weights_distribution = (
            att_blocks.norm(dim=1) / att_blocks.norm(dim=1).sum()
        )
        distribs.append(
            DecoderWeightsDistribution(
                model.cfg.n_heads, [float(x) for x in decoder_weights_distribution]
            )
        )

    return distribs



from dataclasses import dataclass, field
from typing import Any, Callable, List, Literal, Optional

from sae_dashboard.components import (
    ActsHistogramData,
    DecoderWeightsDistribution,
    FeatureTablesData,
    GenericData,
    LogitsHistogramData,
    LogitsTableData,
    SequenceData,
    SequenceMultiGroupData,
)
from sae_dashboard.components_config import (
    ActsHistogramConfig,
    FeatureTablesConfig,
    GenericComponentConfig,
    LogitsHistogramConfig,
    LogitsTableConfig,
    PromptConfig,
    SequencesConfig,
)
from sae_dashboard.html_fns import HTML
from sae_dashboard.layout import SaeVisLayoutConfig


@dataclass
class DFAData:
    dfaValues: List[List[float]] = field(default_factory=list)
    dfaTargetIndex: List[int] = field(default_factory=list)
    dfaMaxValue: float = 0.0


@dataclass
class FeatureData:
    """
    This contains all the data necessary to make the feature-centric visualization, for a single feature. See
    diagram in readme:

        https://github.com/callummcdougall/sae_vis#data_storing_fnspy

    Args:
        feature_idx:    Index of the feature in question (not used within this class's methods, but used elsewhere).
        cfg:            Contains layout parameters which are important in the `get_html` function.

        The other args are the 6 possible components we might have in the feature-centric vis, i.e. this is where we
        store the actual data. Note that one of these arguments is `prompt_data` which is only applicable in the prompt-
        centric view.

    This is used in both the feature-centric and prompt-centric views. In the feature-centric view, a single one
    of these objects creates the HTML for a single feature (i.e. a full screen). In the prompt-centric view, a single
    one of these objects will create one column of the full screen vis.
    """

    feature_tables_data: FeatureTablesData = field(
        default_factory=lambda: FeatureTablesData()
    )
    acts_histogram_data: ActsHistogramData = field(
        default_factory=lambda: ActsHistogramData()
    )
    logits_table_data: LogitsTableData = field(
        default_factory=lambda: LogitsTableData()
    )
    logits_histogram_data: LogitsHistogramData = field(
        default_factory=lambda: LogitsHistogramData()
    )
    sequence_data: SequenceMultiGroupData = field(
        default_factory=lambda: SequenceMultiGroupData()
    )
    prompt_data: SequenceData = field(default_factory=lambda: SequenceData())
    dfa_data: Optional[dict[int, dict[str, Any]]] = None
    decoder_weights_data: Optional[DecoderWeightsDistribution] = None

    def __post_init__(self):
        if self.dfa_data is None:
            self.dfa_data = {}

    def get_component_from_config(self, config: GenericComponentConfig) -> GenericData:
        """
        Given a config object, returns the corresponding data object stored by this instance. For instance, if the input
        is an `FeatureTablesConfig` instance, then this function returns `self.feature_tables_data`.
        """
        CONFIG_CLASS_MAP = {
            FeatureTablesConfig.__name__: self.feature_tables_data,
            ActsHistogramConfig.__name__: self.acts_histogram_data,
            LogitsTableConfig.__name__: self.logits_table_data,
            LogitsHistogramConfig.__name__: self.logits_histogram_data,
            SequencesConfig.__name__: self.sequence_data,
            PromptConfig.__name__: self.prompt_data,
            # Add DFA config here if we create a specific config for it
        }
        config_class_name = config.__class__.__name__
        assert (
            config_class_name in CONFIG_CLASS_MAP
        ), f"Invalid component config: {config_class_name}"
        return CONFIG_CLASS_MAP[config_class_name]

    def _get_html_data_feature_centric(
        self,
        layout: SaeVisLayoutConfig,
        decode_fn: Callable[[int | list[int]], str | list[str]],
    ) -> HTML:
        """
        Returns the HTML object for a single feature-centric view. These are assembled together into the full feature-
        centric view.

        Args:
            decode_fn:  We use this function to decode the token IDs into string tokens.

        Returns:
            html_obj.html_data:
                Contains a dictionary with keys equal to columns, and values equal to the HTML strings. These will be
                turned into grid-column elements, and concatenated.
            html_obj.js_data:
                Contains a dictionary with keys = component names, and values = JavaScript data that will be used by the
                scripts we'll eventually dump in.
        """
        # Create object to store all HTML
        html_obj = HTML()

        # For every column in this feature-centric layout, we add all the components in that column
        for column_idx, column_components in layout.columns.items():
            for component_config in column_components:
                component = self.get_component_from_config(component_config)

                html_obj += component._get_html_data(
                    cfg=component_config,
                    decode_fn=decode_fn,
                    column=column_idx,
                    id_suffix="0",  # we only use this if we have >1 set of histograms, i.e. prompt-centric vis
                )

        return html_obj

    def _get_html_data_prompt_centric(
        self,
        layout: SaeVisLayoutConfig,
        decode_fn: Callable[[int | list[int]], str | list[str]],
        column_idx: int,
        bold_idx: int | Literal["max"],
        title: str,
    ) -> HTML:
        """
        Returns the HTML object for a single column of the prompt-centric view. These are assembled together into a full
        screen of a prompt-centric view, and then they're further assembled together into the full prompt-centric view.

        Args:
            decode_fn:  We use this function to decode the token IDs into string tokens.
            column_idx: This method only gives us a single column (of the prompt-centric vis), so we need to know which
                        column this is (for the JavaScript data).
            bold_idx:   Which index should be bolded in the sequence data. If "max", we default to bolding the max-act
                        token in each sequence.
            title:      The title for this column, which will be used in the JavaScript data.

        Returns:
            html_obj.html_data:
                Contains a dictionary with the single key `str(column_idx)`, representing the single column. This will
                become a single grid-column element, and will get concatenated with others of these.
            html_obj.js_data:
                Contains a dictionary with keys = component names, and values = JavaScript data that will be used by the
                scripts we'll eventually dump in.
        """
        # Create object to store all HTML
        html_obj = HTML()

        # Verify that we only have a single column
        assert layout.columns.keys() == {
            0
        }, f"prompt_centric_layout should only have 1 column, instead found cols {layout.columns.keys()}"
        assert (
            layout.prompt_cfg is not None
        ), "prompt_centric_cfg should include a PromptConfig, but found None"
        if layout.seq_cfg is not None:
            assert (layout.seq_cfg.n_quantiles == 0) or (
                layout.seq_cfg.stack_mode == "stack-all"
            ), "prompt_centric_layout should have stack_mode='stack-all' if n_quantiles > 0, so that it fits in 1 col"

        # Get the maximum color over both the prompt and the sequences
        max_feat_act = max(
            max(self.prompt_data.feat_acts), self.sequence_data.max_feat_act
        )
        max_loss_contribution = max(
            max(self.prompt_data.loss_contribution),
            self.sequence_data.max_loss_contribution,
        )

        # For every component in the single column of this prompt-centric layout, add all the components in that column
        for component_config in layout.columns[0]:
            component = self.get_component_from_config(component_config)

            html_obj += component._get_html_data(
                cfg=component_config,
                decode_fn=decode_fn,
                column=column_idx,
                id_suffix=str(column_idx),
                component_specific_kwargs=dict(  # only used by SequenceData (the prompt)
                    bold_idx=bold_idx,
                    permanent_line=True,
                    hover_above=True,
                    max_feat_act=max_feat_act,
                    max_loss_contribution=max_loss_contribution,
                ),
            )

        # Add the title in JavaScript, and the empty title element in HTML
        html_obj.html_data[column_idx] = (
            f"<div id='column-{column_idx}-title'></div>\n{html_obj.html_data[column_idx]}"
        )
        html_obj.js_data["gridColumnTitlesData"] = {str(column_idx): title}

        return html_obj



import re
from dataclasses import dataclass
from typing import Callable, Dict, List, Sequence, Tuple

import torch
import torch.nn as nn
from jaxtyping import Float, Int
from torch import Tensor
from transformer_lens import HookedTransformer
from transformer_lens.hook_points import HookPoint

DTYPES = {
    "float32": torch.float32,
    "float16": torch.float16,
    "bfloat16": torch.bfloat16,
}


@dataclass
class ActivationConfig:
    primary_hook_point: str
    auxiliary_hook_points: List[str]


class TransformerLensWrapper(nn.Module):
    """
    This class wraps around & extends the TransformerLens model, so that we can make sure things like the forward
    function have a standardized signature.
    """

    def __init__(self, model: HookedTransformer, activation_config: ActivationConfig):
        super().__init__()
        self.model = model
        self.activation_config = activation_config
        self.validate_hook_points()
        self.hook_layer = self.get_layer(self.activation_config.primary_hook_point)

    def validate_hook_points(self):
        """Checks that the hook points are valid and that the model has them"""
        assert (
            self.activation_config.primary_hook_point in self.model.hook_dict
        ), f"Invalid hook point: {self.activation_config.primary_hook_point}"

        for hook_point in self.activation_config.auxiliary_hook_points:
            assert (
                hook_point in self.model.hook_dict
            ), f"Invalid hook point: {hook_point}"

    def get_layer(self, hook_point: str):
        """Get the layer (so we can do the early stopping in our forward pass)"""
        layer_match = re.match(r"blocks\.(\d+)\.", hook_point)
        assert (
            layer_match
        ), f"Error: expecting hook_point to be 'blocks.{{layer}}.{{...}}', but got {hook_point!r}"
        return int(layer_match.group(1))

    def forward(  # type: ignore
        self,
        tokens: Int[Tensor, "batch seq"],
        return_logits: bool = True,
    ) -> Dict[str, Tensor]:
        """Executes a forward pass, collecting specific hook point activations and optionally logit outputs"""
        activation_dict = {}

        def build_act_dict(
            hooks: Sequence[Tuple[str, Callable[[Tensor, HookPoint], None]]]
        ) -> None:
            for hook_point, _ in hooks:
                # The hook functions work by storing data in model's hook context, so we pop them back out

                activation: Tensor = self.model.hook_dict[hook_point].ctx.pop(
                    "activation"
                )
                if "hook_z" in hook_point:
                    activation = activation.flatten(-2, -1)
                activation_dict[hook_point] = activation

        hooks: List[Tuple[str, Callable[[Tensor, HookPoint], None]]] = [
            (self.activation_config.primary_hook_point, self.hook_fn_store_act)
        ] + [
            (point, self.hook_fn_store_act)
            for point in self.activation_config.auxiliary_hook_points
        ]

        output: Tensor = self.model.run_with_hooks(
            tokens, stop_at_layer=self.hook_layer + 1, fwd_hooks=hooks  # type: ignore
        )

        build_act_dict(hooks)

        if return_logits:
            activation_dict["output"] = output
        return activation_dict

    def hook_fn_store_act(self, activation: torch.Tensor, hook: HookPoint):
        hook.ctx["activation"] = activation

    @property
    def tokenizer(self):  # type: ignore
        return self.model.tokenizer

    @property
    def W_U(self):
        return self.model.W_U

    @property
    def W_out(self):
        return self.model.W_out

    @property
    def W_O(self):
        return self.model.W_O


def to_resid_direction(
    direction: Float[Tensor, "feats d_in"], model: TransformerLensWrapper
):
    """
    Takes a direction (eg. in the post-ReLU MLP activations) and returns the corresponding direction in the residual stream.

    Args:
        direction:
            The direction in the activations, i.e. shape (feats, d_in) where d_in could be d_model, d_mlp, etc.
        model:
            The model, which should be a HookedTransformerWrapper or similar.
    """
    # If this SAE was trained on the residual stream or attn/mlp out, then we don't need to do anything
    if (
        "resid" in model.activation_config.primary_hook_point
        or "_out" in model.activation_config.primary_hook_point
    ):
        return direction

    # If it was trained on the MLP layer, then we apply the W_out map
    elif ("pre" in model.activation_config.primary_hook_point) or (
        "post" in model.activation_config.primary_hook_point
    ):
        return direction @ model.W_out[model.hook_layer]

    elif "hook_z" in model.activation_config.primary_hook_point:
        return direction @ model.W_O[model.hook_layer].flatten(0, 1).to(direction.dtype)

    # Others not yet supported
    else:
        raise NotImplementedError(
            "The hook your SAE was trained on isn't yet supported"
        )



import json
import math
import os
from pathlib import Path

import requests
import torch
import typer
from rich import print
from rich.align import Align
from rich.panel import Panel
from sae_lens.analysis.neuronpedia_integration import NanAndInfReplacer
from sae_lens.sae import SAE
from sae_lens.toolkit.pretrained_saes import load_sparsity
from typing_extensions import Annotated

from sae_dashboard.neuronpedia.neuronpedia_runner import (
    NeuronpediaRunner,
    NeuronpediaRunnerConfig,
)
from sae_dashboard.neuronpedia.neuronpedia_runner_config import (
    DEFAULT_SPARSITY_THRESHOLD,
)

OUTPUT_DIR_BASE = Path("./neuronpedia_outputs")
RUN_SETTINGS_FILE = "run_settings.json"

app = typer.Typer(
    add_completion=False,
    no_args_is_help=True,
    help="Tool that generates features (generate) and uploads features (upload) to Neuronpedia.",
)


@app.command()
def generate(
    ctx: typer.Context,
    sae_set: Annotated[
        str,
        typer.Option(
            help="SAE ID to generate features for (must exactly match the one used on Neuronpedia). Example: res-jb",
            prompt="""
What is the SAE ID you want to generate features for?
This was set when you did 'Add SAEs' on Neuronpedia. This must exactly match that ID (including casing).
It's in the format [abbrev hook name]-[abbrev author name], like res-jb.
Enter SAE ID""",
        ),
    ],
    sae_path: Annotated[
        Path,
        typer.Option(
            help="Absolute local path to the SAE directory (with cfg.json, sae_weights.safetensors, sparsity.safetensors).",
            prompt="""
What is the absolute local path to your SAE's directory (with cfg.json, sae_weights.safetensors, sparsity.safetensors)?
Enter path""",
        ),
    ],
    dataset_path: Annotated[
        str,
        typer.Option(
            exists=True,
            dir_okay=True,
            readable=True,
            resolve_path=True,
            help="Path of HuggingFace dataset to use for prompts (default: SAE configured dataset)",
            prompt="""
Override dataset with custom HuggingFace dataset path?
Enter path""",
        ),
    ] = "",
    log_sparsity: Annotated[
        int,
        typer.Option(
            min=-10,
            max=1,
            help="Desired feature log sparsity threshold. Range -10 to 0. Use 1 to skip sparsity check.",
            prompt="""
What is your desired feature log sparsity threshold?
Enter value from -10 to 0 [1 to skip]""",
        ),
    ] = DEFAULT_SPARSITY_THRESHOLD,
    dtype: Annotated[
        str,
        typer.Option(
            help="Override DType",
            prompt="""
Override DType type?
[Enter to use SAE default]""",
        ),
    ] = "",
    feat_per_batch: Annotated[
        int,
        typer.Option(
            min=1,
            max=2048,
            help="Features to generate per batch. More requires more memory.",
            prompt="""
How many features do you want to generate per batch? More requires more memory.
Enter value""",
        ),
    ] = 128,
    n_prompts: Annotated[
        int,
        typer.Option(
            min=1,
            help="[Activation Text Generation] Number of prompts to select from.",
            prompt="""
[Activation Text Generation] How many prompts do you want to select from?
Enter value""",
        ),
    ] = 24576,
    n_context_tokens: Annotated[
        int,
        typer.Option(
            min=0,
            help="[Activation Text Generation] Override the context tokens length.",
            prompt="""
[Activation Text Generation] Override the context tokens length? (0 = Default)
Enter value""",
        ),
    ] = 0,
    n_prompts_in_forward_pass: Annotated[
        int,
        typer.Option(
            min=0,
            help="[Activation Text Generation] Number of prompts per forward pass.",
            prompt="""
[Activation Text Generation] Numbe of prompts per forward pass? (32 = Default)
Enter value""",
        ),
    ] = 32,
    resume_from_batch: Annotated[
        int,
        typer.Option(
            min=0,
            help="Batch number to resume from.",
            prompt="""
Do you want to resume from a specific batch number?
Enter 0 to start from the beginning. Existing batch files will not be overwritten.""",
        ),
    ] = 0,
    end_at_batch: Annotated[
        int,
        typer.Option(
            min=-1,
            help="Batch number to end at.",
            prompt="""
Do you want to end at a specific batch number?
Enter -1 to do all batches. Existing batch files will not be overwritten.""",
        ),
    ] = -1,
):
    """
    This will start a batch job that generates features for Neuronpedia for a specific SAE. To upload those features, use the 'upload' command afterwards.
    """

    print("\nRe-run command with:\n")
    command = "python neuronpedia.py generate"
    for key, value in ctx.params.items():
        command += f" --{key.replace('_', '-')}={value}"  # type: ignore
    print(command + "\n\n")

    # Check arguments
    # if sae_path.is_dir() is not True:
    #     print("Error: SAE path must be a directory.")
    #     raise typer.Abort()
    # if sae_path.joinpath("cfg.json").is_file() is not True:
    #     print("Error: cfg.json file not found in SAE directory.")
    #     raise typer.Abort()
    # if sae_path.joinpath("sae_weights.safetensors").is_file() is not True:
    #     print("Error: sae_weights.safetensors file not found in SAE directory.")
    #     raise typer.Abort()
    # Allow skipping sparsity file
    # if (
    #     log_sparsity != 1
    #     and sae_path.joinpath("sparsity.safetensors").is_file() is not True
    # ):
    #     print("Error: sparsity.safetensors file not found in SAE directory.")
    #     raise typer.Abort()

    sae_path_string = sae_path.as_posix()

    # Load SAE
    device = "cpu"
    if torch.backends.mps.is_available():
        device = "mps"
    elif torch.cuda.is_available():
        device = "cuda"
    try:
        sparse_autoencoder = SAE.load_from_pretrained(sae_path_string, device=device)
    except Exception:
        sparse_autoencoder, _, _ = SAE.from_pretrained(
            sae_set, str(sae_path), device=device
        )
    model_id = sparse_autoencoder.cfg.model_name
    if dtype == "":
        dtype = sparse_autoencoder.cfg.dtype

    if dataset_path == "":
        dataset_path = sparse_autoencoder.cfg.dataset_path

    # make the outputs subdirectory if it doesn't exist, ensure it's not a file
    outputs_subdir = f"{model_id}_{sae_set}_{sparse_autoencoder.cfg.hook_name}"
    outputs_dir = OUTPUT_DIR_BASE.joinpath(outputs_subdir)
    if outputs_dir.exists() and outputs_dir.is_file():
        print(f"Error: Output directory {outputs_dir.as_posix()} exists and is a file.")
        raise typer.Abort()
    outputs_dir.mkdir(parents=True, exist_ok=True)

    # Check if output_dir has a run_settings.json file. If so, load those settings.
    run_settings_path = outputs_dir.joinpath(RUN_SETTINGS_FILE)
    print("\n")
    if run_settings_path.exists() and run_settings_path.is_file():
        # load the json file
        with open(run_settings_path, "r") as f:
            run_settings = json.load(f)
            print(
                f"[yellow]Found existing run_settings.json in {run_settings_path.as_posix()}, checking them."
            )
            if run_settings["log_sparsity"] != log_sparsity:
                print(
                    f"[red]Error: log_sparsity in {run_settings_path.as_posix()} doesn't match the current log_sparsity:\n{run_settings['log_sparsity']} vs {log_sparsity}"
                )
                raise typer.Abort()
            if run_settings["sae_set"] != sae_set:
                print(
                    f"[red]Error: sae_set in {run_settings_path.as_posix()} doesn't match the current sae_set:\n{run_settings['sae_set']} vs {sae_set}"
                )
                raise typer.Abort()
            if run_settings["dtype"] != dtype:
                print(
                    f"[red]Error: dtype in {run_settings_path.as_posix()} doesn't match the current dtype:\n{run_settings['dtype']} vs {dtype}"
                )
                raise typer.Abort()
            if run_settings["sae_path"] != sae_path_string:
                print(
                    f"[red]Error: sae_path in {run_settings_path.as_posix()} doesn't match the current sae_path:\n{run_settings['sae_path']} vs {sae_path_string}"
                )
                raise typer.Abort()
            if run_settings["n_prompts"] != n_prompts:
                print(
                    f"[red]Error: n_prompts in {run_settings_path.as_posix()} doesn't match the current n_prompts:\n{run_settings['n_prompts']} vs {n_prompts}"
                )
                raise typer.Abort()
            if run_settings["n_context_tokens"] != n_context_tokens:
                print(
                    f"[red]Error: n_context_tokens in {run_settings_path.as_posix()} doesn't match the current n_context_tokens:\n{run_settings['n_context_tokens']} vs {n_context_tokens}"
                )
                raise typer.Abort()
            if run_settings["feat_per_batch"] != feat_per_batch:
                print(
                    f"[red]Error: feat_per_batch in {run_settings_path.as_posix()} doesn't match the current feat_per_batch:\n{run_settings['feat_per_batch']} vs {feat_per_batch}"
                )
                raise typer.Abort()
            print("[green]All settings match, using existing run_settings.json.")
    else:
        print(f"[green]Creating run_settings.json in {run_settings_path.as_posix()}.")
        run_settings = {
            "sae_set": sae_set,
            "sae_path": sae_path_string,
            "log_sparsity": log_sparsity,
            "dtype": dtype,
            "n_prompts": n_prompts,
            "n_context_tokens": n_context_tokens,
            "feat_per_batch": feat_per_batch,
        }
        with open(run_settings_path, "w") as f:
            json.dump(run_settings, f, indent=4)

    if log_sparsity == 1:
        num_alive = sparse_autoencoder.cfg.d_sae
        num_dead = 0
    else:
        sparsity = load_sparsity(sae_path_string)
        # convert sparsity to logged sparsity if it's not
        # TODO: standardize the sparsity file format
        if len(sparsity) > 0 and sparsity[0] >= 0:
            sparsity = torch.log10(sparsity + 1e-10)
        sparsity = sparsity.to(device)
        alive_indexes = (sparsity > log_sparsity).nonzero(as_tuple=True)[0].tolist()
        num_alive = len(alive_indexes)
        num_dead = sparse_autoencoder.cfg.d_sae - num_alive

    num_batches = math.ceil(sparse_autoencoder.cfg.d_sae / feat_per_batch)
    if end_at_batch >= num_batches:
        print(
            f"[red]Error: end_at_batch {end_at_batch} should not be >= num_batches {num_batches}"
        )
        raise typer.Abort()

    print("\n")
    print(
        Align.center(
            Panel.fit(
                f"""
[white]SAE Path: [green]{sae_path.as_posix()}
[white]Dataset Path: [green]{dataset_path}
[white]Model ID: [green]{model_id}
[white]Hook Point: [green]{sparse_autoencoder.cfg.hook_name}
[white]DType: [green]{dtype}
[white]Using Device: [green]{device}
""",
                title="SAE Info",
            )
        )
    )
    print(
        Align.center(
            Panel.fit(
                f"""
[white]Total Features: [green]{sparse_autoencoder.cfg.d_sae}
[white]Log Sparsity Threshold: [green]{log_sparsity}
[white]Alive Features: [green]{num_alive}
[white]Dead Features: [red]{num_dead}
[white]Features per Batch: [green]{feat_per_batch}
[white]Number of Batches: [green]{num_batches}
{resume_from_batch != 0 and f"[white]Resuming from Batch: [green]{resume_from_batch}" or ""}
{end_at_batch != -1 and f"[white]Ending at Batch: [green]{end_at_batch}" or ""}
""",
                title="Number of Features",
            )
        )
    )
    print(
        Align.center(
            Panel.fit(
                f"""
[white]Dataset: [green]{dataset_path}
[white]Prompts to Sample From: [green]{n_prompts}
[white]Context Token Length: [green]{n_context_tokens if n_context_tokens != 0 else 0}
[white]Prompts per Forward Pass: [green]{n_prompts_in_forward_pass}
""",
                title="Activation Text Settings",
            )
        )
    )
    print(
        Align.center(
            Panel.fit(
                f"""
[green]{outputs_dir.absolute().as_posix()}
""",
                title="Output Directory",
            )
        )
    )

    print(
        Align.center(
            "\n========== [yellow]Starting batch feature generations...[/yellow] ==========\n"
        )
    )

    print("\n\nRe-run with command:\n")
    print(command + "\n\n")

    # run the command
    cfg = NeuronpediaRunnerConfig(
        sae_set=sae_set,
        sae_path=str(sae_path),  # .absolute().as_posix(),
        huggingface_dataset_path=dataset_path,
        sae_dtype=dtype,
        model_dtype=dtype,
        outputs_dir=outputs_dir.absolute().as_posix(),
        sparsity_threshold=log_sparsity,
        n_prompts_total=n_prompts,
        n_tokens_in_prompt=n_context_tokens if n_context_tokens != 0 else 128,
        n_prompts_in_forward_pass=n_prompts_in_forward_pass,
        n_features_at_a_time=feat_per_batch,
        start_batch=resume_from_batch,
        end_batch=end_at_batch if end_at_batch != -1 else num_batches,
    )

    runner = NeuronpediaRunner(cfg)
    runner.run()

    print(
        Align.center(
            Panel(
                f"""
Your Features Are In: [green]{outputs_dir.absolute().as_posix()}
Use [yellow]'neuronpedia.py upload'[/yellow] to upload your features to Neuronpedia.
""",
                title="Generation Complete",
            )
        )
    )


@app.command()
def upload(
    outputs_dir: Annotated[
        Path,
        typer.Option(
            exists=True,
            dir_okay=True,
            readable=True,
            resolve_path=True,
            prompt="What is the absolute local file path to the feature outputs directory?",
        ),
    ],
    api_key: Annotated[
        str,
        typer.Option(
            prompt="""Your Neuronpedia API key? (input will be hidden)""",
            hide_input=True,
        ),
    ],
    host: Annotated[
        str,
        typer.Option(
            prompt="""Host to upload to? (Default: http://localhost:3000)""",
        ),
    ] = "http://localhost:3000",
    resume: Annotated[
        int,
        typer.Option(
            prompt="""Resume from batch? (Default: 0)""",
        ),
    ] = 0,
    upload_dead_stubs: Annotated[
        bool,
        typer.Option(
            prompt="""Upload dead stubs? (Default: True)""",
        ),
    ] = True,
):
    """
    This will upload features that were generated to Neuronpedia. It currently only works if you have admin access to a Neuronpedia instance via localhost:3000.
    """

    if upload_dead_stubs:
        skipped_path = os.path.join(outputs_dir, "skipped_indexes.json")
        f = open(skipped_path, "r")
        data = json.load(f)
        url = host + "/api/local/upload-skipped-features"
        result = requests.post(
            url,
            json=data,
            headers={"x-api-key": api_key},
        )
        if result.status_code != 200:
            raise Exception(f"Upload failed with status code: {result.status_code}")

    files_to_upload = list(outputs_dir.glob("batch-*.json"))

    # filter files where batch-[number].json the number is >= resume
    files_to_upload = [
        file_path
        for file_path in files_to_upload
        if int(file_path.stem.split("-")[1]) >= resume
    ]

    # sort files by batch number
    files_to_upload.sort(key=lambda x: int(x.stem.split("-")[1]))

    print("\n")
    # Upload alive features
    for file_path in files_to_upload:
        print("===== Uploading file: " + os.path.basename(file_path))
        f = open(file_path, "r")
        data = json.load(f, parse_constant=NanAndInfReplacer)

        url = host + "/api/local/upload-features"
        result = requests.post(
            url,
            json=data,
            headers={"x-api-key": api_key},
        )
        if result.status_code != 200:
            raise Exception(f"Upload failed with status code: {result.status_code}")

    print(
        Align.center(
            Panel(
                f"""
{len(files_to_upload)} batch files uploaded to Neuronpedia.
""",
                title="Uploads Complete",
            )
        )
    )


if __name__ == "__main__":
    app()
